import pathlib
from abc import abstractmethod
from dataclasses import dataclass
from typing import List
import inspect

import click
from jinja2 import Environment, PackageLoader, select_autoescape, TemplateNotFound

from coworks import TechMicroService
from coworks.config import CORSConfig
from coworks.cws.command import CwsCommand
from coworks.cws.error import CwsCommandError

DEFAULT_STEP = 'update'


class CwsWriterError(CwsCommandError):
    ...


class CwsWriter(CwsCommand):

    def __init__(self, app=None, *, name):
        super().__init__(app, name=name)

    @property
    def options(self):
        return [
            *super().options,
            click.option('-o', '--output'),
            click.option('--step', default=DEFAULT_STEP),
            click.option('--config'),
            click.option('--debug/--no-debug', default=False, help='Print debug logs to stderr.')
        ]

    def _execute(self, **options):
        self._export_header(**options)
        self._export_content(**options)

        print('', file=self.output, flush=True)

    def _export_header(self, **options):
        ...

    @abstractmethod
    def _export_content(self, **options):
        """ Main export function.
        :param options: Command optons.
        :return: None.

        Abstract method which must be redefined in any subclass. The content should be written in self.output.
        """

    def _format(self, content):
        return content


class CwsTemplateWriter(CwsWriter):
    """Writer with  jinja templating."""

    def __init__(self, app=None, *, name='export', data=None, template_filenames=None, env=None):
        super().__init__(app, name=name)
        self.data = data or {}
        self.template_filenames = template_filenames or self.default_template_filenames
        self.env = env or Environment(
            loader=PackageLoader("coworks.cws.writer"),
            autoescape=select_autoescape(['html', 'xml'])
        )

    @property
    @abstractmethod
    def default_template_filenames(self):
        """Must be redefined to set template file for writing."""

    def _export_content(self, *, project_dir, module, service, workspace, **options):
        module_path = module.split('.')

        # Get parameters for execution
        try:
            config = next(
                (app_config for app_config in self.app.configs if app_config.workspace == workspace)
            )
        except StopIteration:
            raise CwsCommandError("A workspace is mandatory in the python configuration for deploying.\n")

        environment_variable_files = [p.as_posix() for p in
                                      config.existing_environment_variables_files(project_dir)]
        data = {
            'writer': self,
            'project_dir': project_dir,
            'source_file': pathlib.PurePath(project_dir, *module_path),
            'module': module,
            'module_path': pathlib.PurePath(*module_path),
            'module_dir': pathlib.PurePath(*module_path[:-1]),
            'module_file': module_path[-1],
            'handler': service,
            'app': self.app,
            'ms_name': self.app.name,
            'workspace': workspace,
            'app_config': config,
            'environment_variable_files': environment_variable_files,
            'sfn_name': options.get('sfn_name'),
            'account_number': options.get('account_number'),
            'description': inspect.getdoc(self.app),
            **options
        }
        data.update(self.data)
        try:
            for template_filename in self.template_filenames:
                template = self.env.get_template(template_filename)
                print(self._format(template.render(**data)), file=self.output)
        except TemplateNotFound as e:
            raise CwsWriterError(f"Cannot find template {str(e)}")
        except Exception as e:
            raise CwsWriterError(e)


UID_SEP = '_'


@dataclass
class TerraformEntry:
    app: TechMicroService
    parent_uid: str
    path: str
    methods: List[str]
    cors: CORSConfig

    @property
    def uid(self):
        def remove_brackets(path):
            return f"{path.replace('{', '').replace('}', '')}"

        if self.path is None:
            return UID_SEP

        last = remove_brackets(self.path)
        return f"{self.parent_uid}{UID_SEP}{last}" if self.parent_uid else last

    @property
    def is_root(self):
        return self.path is None

    @property
    def parent_is_root(self):
        return self.parent_uid == UID_SEP

    def __repr__(self):
        return f"{self.uid}:{self.methods}"


class CwsTerraformWriter(CwsTemplateWriter):

    def __init__(self, app=None, *, name='terraform', data=None, **kwargs):

        data = data or {
            'layer_zip_file': 'layer.zip',
        }
        super().__init__(app, name=name, data=data, **kwargs)

    @property
    def default_template_filenames(self):
        return ['terraform.j2']

    def _export_header(self, **options):
        print("// Do NOT edit this file as it is auto-generated by cws\n", file=self.output)

    @property
    def entries(self):
        """Returns the list of flatten path (prev, last, keys)."""
        all_pathes_id = {}

        def add_entry(previous, last, meth):
            entry = TerraformEntry(self.app, previous, last, meth, self.app.config.cors)
            uid = entry.uid
            if uid not in all_pathes_id:
                all_pathes_id[uid] = entry
            if all_pathes_id[uid].methods is None:
                all_pathes_id[uid].methods = meth
            return uid

        for route, methods in self.app.routes.items():
            previous_uid = UID_SEP
            splited_route = route[1:].split('/')

            # special root case
            if splited_route == ['']:
                add_entry(None, None, methods.keys())
                continue

            # creates intermediate resources
            last_path = splited_route[-1:][0]
            for prev in splited_route[:-1]:
                previous_uid = add_entry(previous_uid, prev, None)

            # set entryes keys for last entry
            add_entry(previous_uid, last_path, methods.keys())

        return all_pathes_id

    def _execute(self, **options):
        if options['debug']:
            print(f"Generate terraform file: {options['output']}")
        super()._execute(**options)



import csv
import io
import json
from datetime import datetime
from typing import Union, List

from ..coworks import TechMicroService
from coworks.mixins import Boto3Mixin, AwsS3Session


class CSVMicroService(TechMicroService, Boto3Mixin):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.aws_s3_session = AwsS3Session()

    def upload_to_s3(self, file_obj, bucket, expiration=3600):
        """ Upload a file to s3 and return a presigned url to download it """
        try:
            response = self.aws_s3_session.client.upload_fileobj(
                file_obj, bucket, file_obj.name, ExtraArgs={'ContentType': 'text/csv'})
        except Exception as e:
            print(e)
        try:
            response = self.aws_s3_session.client.generate_presigned_url(
                'get_object', Params={'Bucket': bucket, 'Key': file_obj.name}, ExpiresIn=expiration)
            return response
        except Exception as e:
            print(e)

    def post_format(self, content: str = "", title: Union[bool, List[str]] = True,
                    remove_rows: List[int] = None, remove_columns: List[int] = None,
                    delimiter: str = ',', upload_to_s3_bucket: str = None, s3_bucket_folder: str = None):
        """Format JSON list to CSV content."""

        remove_rows = remove_rows if remove_rows is not None else []
        remove_columns = remove_columns if remove_columns is not None else []
        if type(remove_rows) is not list:
            raise EnvironmentError("remove_rows parameter must be a list")
        if type(remove_columns) is not list:
            raise EnvironmentError("remove_columns parameter must be a list")

        rows = json.loads(content) if type(content) == str else content
        output = io.StringIO()
        writer = csv.writer(output, quoting=csv.QUOTE_NONNUMERIC, delimiter=delimiter)

        if title is True:  # takes title from keys of first line
            line = [key for idx, key in enumerate(rows[0]) if idx not in remove_columns]
            writer.writerow(line)
        elif type(title) is list:
            line = [col for idx, col in enumerate(title) if idx not in remove_columns]
            writer.writerow(line)
        for index, row in enumerate(rows):
            if index not in remove_rows:
                line = [col for idx, col in enumerate(row.values()) if idx not in remove_columns]
                writer.writerow(line)

        res = output.getvalue()

        if upload_to_s3_bucket:
            file_obj = io.BytesIO(res.encode('utf-8'))
            if s3_bucket_folder:
                file_obj.name = f"{s3_bucket_folder}/{datetime.now()}.csv"
            else:
                file_obj.name = f"{datetime.now()}.csv"
            return self.upload_to_s3(file_obj, upload_to_s3_bucket)
        else:
            return res



from coworks import BizFactory, Every
from coworks.cws.sfn import StepFunctionWriter

biz = BizFactory(ms_name='biz')
reactor1 = biz.create('small', 'often', Every(1, Every.DAYS))
reactor2 = biz.create('complete', 'often', Every(1, Every.DAYS))
reactor3 = biz.create('complete', 'rarely', Every(365, Every.DAYS))
StepFunctionWriter(biz)



import io
import re

from coworks.config import Config, CORSConfig
from coworks.cws.writer import CwsTerraformWriter
from tests.src.coworks.tech_ms import SimpleMS


class TestClass:

    def test_export_terraform(self):
        simple = SimpleMS()
        writer = CwsTerraformWriter(simple)
        output = io.StringIO()
        self.export_cmd(simple, output)
        output.seek(0)
        print(output.read())
        output.seek(0)
        assert len(re.sub(r"\s", "", output.read())) == 8449
        assert len(writer.entries) == 8

        assert writer.entries['_'].parent_uid is None
        assert writer.entries['_'].is_root
        assert writer.entries['_'].path is None
        assert 'GET' in writer.entries['_'].methods
        assert 'POST' not in writer.entries['_'].methods

        assert writer.entries['__content'].parent_uid == '_'
        assert not writer.entries['__content'].is_root
        assert writer.entries['__content'].parent_is_root
        assert writer.entries['__content'].path == 'content'
        assert 'GET' in writer.entries['__content'].methods
        assert 'POST' in writer.entries['__content'].methods

        assert writer.entries['__extended'].methods is None

        assert not writer.entries['__extended_content'].is_root
        assert not writer.entries['__extended_content'].parent_is_root
        assert 'GET' in writer.entries['__extended_content'].methods

    def test_export_terraform_with_cors(self):
        config = Config(cors=CORSConfig(allow_origin='www.test.fr'))
        simple = SimpleMS(configs=config)
        CwsTerraformWriter(simple)
        output = io.StringIO()
        self.export_cmd(simple, output)
        output.seek(0)
        print(output.read())
        output.seek(0)
        assert len(re.sub(r"\s", "", output.read())) == 19606

    @staticmethod
    def export_cmd(simple, output):
        options = {'project_dir': '.', 'module': "test", 'workspace': 'dev', 'step': 'update'}
        simple.execute('terraform', output=output, **options)



import cgi
import inspect
import io
import json
import os
import traceback
import urllib
from functools import update_wrapper, partial

import boto3
from aws_xray_sdk.core import xray_recorder
from botocore.exceptions import BotoCoreError
from chalice import AuthResponse, BadRequestError, Response
from requests_toolbelt.multipart import MultipartDecoder

from coworks.error import CwsError
from .utils import HTTP_METHODS, class_auth_methods, class_http_methods, trim_underscores, make_absolute


class EntryPoint:

    def __init__(self, component, auth, fun):
        self.component = component
        self.auth = auth
        self.fun = fun


class Entry(dict):

    @property
    def authorizer(self):
        for method in HTTP_METHODS:
            m = self.get(method.upper())
            if m:
                return m.auth

    def call_get(self, *args, **kwargs):
        method = self["GET"]
        return method.fun(method.component, *args, **kwargs)

    def call_post(self, *args, **kwargs):
        method = self["POST"]
        return method.fun(method.component, *args, **kwargs)


class CoworksMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.debug = kwargs.pop('debug', False)

        # A list of functions that will be called at the first activation.
        # To register a function, use the :meth:`before_first_request` decorator.
        self.before_first_activation_funcs = []

        # A list of functions that will be called at the beginning of each activation.
        # To register a function, use the :meth:`before_activation` decorator.
        self.before_activation_funcs = []

        # A list of functions that will be called after each activation.
        # To register a function, use the :meth:`after_activation` decorator.
        self.after_activation_funcs = []

        # A list of functions that will be called in case of exception.
        # To register a function, use the :meth:`handle_exception` decorator.
        self.handle_exception_funcs = []

        self.aws_s3_sfn_data_session = AwsS3Session(env_var_access_key="AWS_RUN_ACCESS_KEY_ID",
                                                    env_var_secret_key="AWS_RUN_SECRET_KEY",
                                                    env_var_region="AWS_RUN_REGION")
        self.aws_s3_form_data_session = AwsS3Session(env_var_access_key="AWS_FORM_DATA_ACCESS_KEY_ID",
                                                     env_var_secret_key="AWS_FORM_DATA_SECRET_KEY",
                                                     env_var_region="AWS_FORM_DATA_REGION")

    def before_first_activation(self, f):
        """Registers a function to be run before the first activation of the microservice.

        May be used as a decorator.

        The function will be called with event and context positional arguments and its return value is ignored.
        """

        self.before_first_activation_funcs.append(f)
        return f

    def before_activation(self, f):
        """Registers a function to called before each activation of the microservice.
        :param f:  Function added to the list.
        :return: None.

        May be used as a decorator.

        The function will be called with event and context positional arguments and its return value is ignored.
        """

        self.before_activation_funcs.append(f)
        return f

    def after_activation(self, f):
        """Registers a function to be called after each activation of the microservice.

        May be used as a decorator.

        The function will be called with the response and its return value is the final response.
        """

        self.after_activation_funcs.append(f)
        return f

    def handle_exception(self, f):
        """Registers a function to be called in case of exception.

        May be used as a decorator.

        The function will be called with the exception and eventually a response to return.
        """

        self.handle_exception_funcs.append(f)
        return f

    def _init_routes(self, app, *, url_prefix='', authorizer=None, hide_routes=False):
        # External authorizer has priority (forced)
        if authorizer is None:
            auth_fun = app.config.auth if app.config.auth else class_auth_methods(self)
            auth = self._create_auth_proxy(auth_fun) if auth_fun else None
        else:
            auth = authorizer

        # Adds entrypoints
        methods = class_http_methods(self)
        for method, func in methods:

            # Get function's route
            if func.__name__ == method:
                route = f"{url_prefix}"
            else:
                name = func.__name__[len(method) + 1:]
                name = trim_underscores(name)  # to allow several functions with same route but different args
                name = name.replace('_', '/')
                route = f"{url_prefix}/{name}" if url_prefix else f"{name}"

            # Get parameters
            args = inspect.getfullargspec(func).args[1:]
            defaults = inspect.getfullargspec(func).defaults
            varkw = inspect.getfullargspec(func).varkw
            if defaults:
                len_defaults = len(defaults)
                for index, arg in enumerate(args[:-len_defaults]):
                    route = route + f"/{{_{index}}}" if route else f"{{_{index}}}"
                kwarg_keys = args[-len_defaults:]
            else:
                for index, arg in enumerate(args):
                    route = route + f"/{{_{index}}}" if route else f"{{_{index}}}"
                kwarg_keys = {}

            proxy = self._create_rest_proxy(func, kwarg_keys, args, varkw)

            # complete all entries
            route = make_absolute(route)
            app.entries[route][method.upper()] = EntryPoint(self, auth, func)
            if not hide_routes and not getattr(func, '__cws_hidden', False):
                app.route(f"{route}", methods=[method.upper()], authorizer=auth, cors=app.config.cors,
                          content_types=list(app.config.content_type))(proxy)

    def _create_auth_proxy(self, auth_method):

        def proxy(auth_activation):
            subsegment = xray_recorder.current_subsegment()
            try:
                auth = auth_method(self, auth_activation)
                if subsegment:
                    subsegment.put_metadata('result', auth)
            except Exception as e:
                self.logger.info(f"Exception : {str(e)}")
                traceback.print_exc()
                if subsegment:
                    subsegment.add_exception(e, traceback.extract_stack())
                raise BadRequestError(str(e))

            if type(auth) is bool:
                if auth:
                    return AuthResponse(routes=['*'], principal_id='user')
                return AuthResponse(routes=[], principal_id='user')
            elif type(auth) is list:
                return AuthResponse(routes=auth, principal_id='user')
            return auth

        proxy = update_wrapper(proxy, auth_method)
        proxy.__name__ = 'app'
        return self.authorizer(name='auth')(proxy)

    def _create_rest_proxy(self, func, kwarg_keys, args, varkw):
        original_app_class = self.__class__

        def proxy(**kws):
            try:
                # Renames positional parameters (index added in label)
                kwargs = {}
                for kw, value in kws.items():
                    param = args[int(kw[1:])]
                    kwargs[param] = value

                # Adds kwargs parameters
                def check_param_expected_in_lambda(param_name):
                    """Alerts when more parameters than expected are defined in request."""
                    if param_name not in kwarg_keys and varkw is None:
                        raise BadRequestError(f"TypeError: got an unexpected keyword argument '{param_name}'")

                def add_param(param_name, param_value):
                    check_param_expected_in_lambda(param_name)
                    if param_name in params:
                        if isinstance(params[param_name], list):
                            params[param_name].append(param_value)
                        else:
                            params[param_name] = [params[param_name], param_value]
                    else:
                        params[param_name] = param_value

                req = self.current_request
                if kwarg_keys or varkw:
                    params = {}
                    if req.raw_body:  # POST request
                        content_type = req.headers['content-type']
                        if content_type.startswith('multipart/form-data'):
                            try:
                                multipart_decoder = MultipartDecoder(req.raw_body, content_type)
                                for part in multipart_decoder.parts:
                                    name, content = self._get_multipart_content(part)
                                    add_param(name, content)
                            except Exception as e:
                                raise CwsError(str(e))
                            kwargs = dict(**kwargs, **params)
                        elif content_type.startswith('application/x-www-form-urlencoded'):
                            params = urllib.parse.parse_qs(req.raw_body.decode("utf-8"))
                            kwargs = dict(**kwargs, **params)
                        elif content_type.startswith('application/json'):
                            if hasattr(req.json_body, 'items'):
                                params = {}
                                for k, v in req.json_body.items():
                                    add_param(k, v)
                                kwargs = dict(**kwargs, **params)
                            else:
                                kwargs[kwarg_keys[0]] = req.json_body
                        elif content_type.startswith('text/plain'):
                            kwargs[kwarg_keys[0]] = req.json_body
                        else:
                            BadRequestError(f"Cannot manage content type {content_type} for {self}")

                    else:  # GET request

                        # adds parameters from qurey parameters
                        for k in req.query_params or []:
                            value = req.query_params.getlist(k)
                            add_param(k, value if len(value) > 1 else value[0])
                        kwargs = dict(**kwargs, **params)
                else:
                    if not args and (req.raw_body or req.query_params):
                        raise BadRequestError(f"TypeError: got an unexpected arguments")

                # chalice is changing class for local server for threading reason (why not mixin..?)
                self_class = self.__class__
                if self_class != original_app_class:
                    self.__class__ = original_app_class

                resp = func(self, **kwargs)
                self.__class__ = self_class
                return _convert_response(resp)

            except Exception as e:
                subsegment = xray_recorder.current_subsegment()
                if subsegment:
                    subsegment.add_error_flag()
                raise

        proxy = update_wrapper(proxy, func)
        proxy.__cws_func__ = update_wrapper(partial(func, self), func)
        return proxy

    def _get_multipart_content(self, part):
        headers = {k.decode('utf-8'): cgi.parse_header(v.decode('utf-8')) for k, v in part.headers.items()}
        content = part.content
        _, content_disposition_params = headers['Content-Disposition']
        part_content_type, _ = headers.get('Content-Type', (None, None))
        name = content_disposition_params['name']

        # content in a text or json value
        if 'filename' not in content_disposition_params:
            if part_content_type == 'application/json':
                return name, self._get_data_on_s3(json.loads(content.decode('utf-8')))
            return name, self._get_data_on_s3(content.decode('utf-8'))

        # content in a file (s3 or plain text)
        if part_content_type == 'text/s3':
            pathes = content.decode('utf-8').split('/', 1)
            try:
                s3_object = self.aws_s3_form_data_session.client.get_object(Bucket=pathes[0], Key=pathes[1])
            except BotoCoreError:
                return CwsError(f"Bucket={pathes[0]} Key={pathes[1]} not found on s3")
            file = io.BytesIO(s3_object['Body'].read())
            mime_type = s3_object['ContentType']
        else:
            file = io.BytesIO(content)
            mime_type = part_content_type
        file.name = content_disposition_params['filename']

        return name, FileParam(file, mime_type)

    def _set_multipart_content(self, form_data):
        def encode_part(_part):
            if type(_part) is str:
                return None, _part, 'text/plain'

            if 'mime_type' in _part:
                mime_type = _part.get('mime_type')
            elif 'json' in _part:
                _part['content'] = _part.get('json')
                mime_type = 'application/json'
            elif 's3' in _part:
                path = _part.get('s3')
                _part['filename'] = path.split('/')[-1]
                _part['path'] = path
                mime_type = 'text/s3'
            else:
                mime_type = 'text/plain'

            filename = _part.get('filename')
            if mime_type == 'text/plain':
                content = _part.get('content')
                return filename, content, mime_type
            elif mime_type == 'application/json':
                content = _part.get('content')
                return filename, json.dumps(content), mime_type
            elif mime_type == 'text/s3':
                path = _part.get('path')
                return filename, path, mime_type
            else:
                raise BadRequestError(f"Undefined mime type {mime_type}")

        parts = []
        for name, part in form_data.items():
            if type(part) is list:
                parts.extend([(name, encode_part(p)) for p in part])
            else:
                parts.append((name, encode_part(part)))
        return parts

    def _set_data_on_s3(self, data):
        """Saves value on S3 temporary file if content is too big."""

        def set_on_s3(value):
            s3_client = self.aws_s3_sfn_data_session.client
            context = self.lambda_context
            key = f"tmp/{context.aws_request_id}"
            tags = f"Name={context.function_name}"
            s3_client.put_object(Bucket="coworks-microservice", Key=key, Body=value, Tagging=tags)
            return f"$${key}$$"

        if type(data) == str:
            return set_on_s3(data) if len(data) > 1000 else data
        if type(data) == list:
            return [self._set_data_on_s3(v) for v in data]
        if type(data) == dict:
            for k, v in data.items():
                data[k] = self._set_data_on_s3(v)
        return data

    def _get_data_on_s3(self, data):
        """Retrieves value from S3 temporary file (content too big)."""

        def get_on_s3(value):
            s3_client = self.aws_s3_sfn_data_session.client
            s3_object = s3_client.get_object(Bucket="coworks-microservice", Key=value[2:-2])
            return s3_object['Body'].read().decode("utf-8")

        if type(data) == str:
            return get_on_s3(data) if data.startswith('$$') and data.endswith('$$') else data
        if type(data) == list:
            return [self._get_data_on_s3(v) for v in data]
        if type(data) == dict:
            for k, v in data.items():
                data[k] = self._get_data_on_s3(v)
        return data


def _convert_response(resp):
    if type(resp) is tuple:
        if len(resp) == 2:
            if type(resp[1]) is int:
                return Response(body=resp[0], status_code=resp[1])
            elif type(resp[1]) is dict:
                return Response(body=resp[0], status_code=200, headers=resp[1])
            else:
                raise BadRequestError("Internal error (wrong result type)")
        else:
            return Response(body=resp[0], status_code=resp[1], headers=resp[2])

    elif not isinstance(resp, Response):
        return Response(body=resp)

    return resp


class Boto3Mixin:

    def __init__(self, service='s3', profile_name=None, env_var_access_key='aws_access_key_id',
                 env_var_secret_key='aws_secret_access_key', env_var_region='aws_region',
                 **kwargs):
        super().__init__(**kwargs)
        self.__session__ = self.__client__ = None
        self.__service = service
        self.__env_var_access_key = env_var_access_key
        self.__env_var_secret_key = env_var_secret_key
        self.__env_var_region = env_var_region

        self.__profile_name = profile_name

    @property
    def aws_access_key(self):
        value = os.getenv(self.__env_var_access_key)
        if not value:
            raise EnvironmentError(f"{self.__env_var_access_key} not defined in environment")
        return value

    @property
    def aws_secret_access_key(self):
        value = os.getenv(self.__env_var_secret_key)
        if not value:
            raise EnvironmentError(f"{self.__env_var_secret_key} not defined in environment")
        return value

    @property
    def region_name(self):
        value = os.getenv(self.__env_var_region)
        if not value:
            raise EnvironmentError(f"{self.__env_var_region} not defined in environment")
        return value

    @property
    def client(self):
        if self.__client__ is None:
            self.__client__ = self.__session.client(self.__service)
        return self.__client__

    @property
    def __session(self):
        if self.__session__ is None:
            region_name = self.region_name
            if self.__profile_name is not None:
                try:
                    self.__session__ = boto3.Session(profile_name=self.__profile_name, region_name=region_name)
                except Exception:
                    raise CwsError(f"Cannot create session for profile {self.__profile_name} and region {region_name}")
            else:
                access_key = self.aws_access_key
                secret_key = self.aws_secret_access_key
                try:
                    self.__session__ = boto3.Session(access_key, secret_key, region_name=region_name)
                except Exception:
                    raise CwsError(
                        f"Cannot create session for key {access_key}, secret {secret_key} and region {region_name}")
        return self.__session__


class FileParam:

    def __init__(self, file, mime_type):
        self.file = file
        self.mime_type = mime_type

    def __repr__(self):
        if self.mime_type:
            return f'FileParam({self.file.name}, {self.mime_type})'
        return f'FileParam({self.file.name})'


class AwsS3Session(Boto3Mixin):

    def __init__(self, **kwargs):
        super().__init__(service='s3', **kwargs)


class AwsSFNSession(Boto3Mixin):

    def __init__(self, **kwargs):
        super().__init__(service='stepfunctions', **kwargs)



import os
from unittest.mock import Mock

import pytest

from coworks import BizFactory, Every, At
from coworks.config import Config


class TestClass:

    def test_biz_reactor(self):
        fact = BizFactory('step_function')
        fact.create('test', Every(5, Every.MINUTES))
        assert len(fact.trigger_sources) == 1
        assert fact.trigger_sources[0]['name'] == 'step_function-test'
        assert fact.trigger_sources[0]['source'] == 'every'
        assert fact.trigger_sources[0]['value'] == 'rate(5 minutes)'
        with pytest.raises(Exception) as execinfo:
            fact.create('test', At(5, 10))
        assert len(fact.trigger_sources) == 1
        assert str(execinfo.value.args[0]) == 'BadRequestError: Biz microservice test already defined for step_function'
        fact.create('test2', At(5, 10, day_of_week='*'))
        assert len(fact.trigger_sources) == 2
        assert fact.trigger_sources[1]['name'] == 'step_function-test2'
        assert fact.trigger_sources[1]['source'] == 'at'
        assert fact.trigger_sources[1]['value'] == 'cron(5 10 None None * None)'

    def test_biz_factory(self):
        fact = BizFactory('sfn_name1', configs=[Config(workspace='test')])
        fact.__sfn_client__ = Mock()
        fact.__sfn_arn__ = "arn1"
        os.environ['WORKSPACE'] = 'test'

        test_config_1 = Config(
            workspace="test",
            data={"key1": "value1"}
        )

        test_config_2 = Config(
            workspace="test",
            data={"key2": "value2"}
        )

        biz_1 = fact.create('every1', Every(5, Every.MINUTES), configs=[test_config_1])
        assert len(fact.biz) == 1
        assert len(fact.trigger_sources) == 1
        assert 'every1' in fact.biz
        biz_2 = fact.create('at2', At('0/10', '*', day_of_month='?', day_of_week='MON-FRI'), configs=[test_config_2])
        assert len(fact.biz) == 2
        assert len(fact.trigger_sources) == 2
        assert 'at2' in fact.biz

        biz_1({'detail-type': 'Scheduled Event',
               'resources': ['arn:aws:events:eu-west-1:123456789:rule/sfn_name1-every1']}, {})
        fact.sfn_client.start_execution.assert_called_once_with(input='{"key1": "value1"}', stateMachineArn='arn1')
        biz_2({'detail-type': 'Scheduled Event',
               'resources': ['arn:aws:events:eu-west-1:123456789:rule/sfn_name1-at2']}, {})
        fact.sfn_client.start_execution.assert_called_with(input='{"key2": "value2"}', stateMachineArn='arn1')

    def test_biz_data(self):
        os.environ['WORKSPACE'] = 'test'
        fact = BizFactory(sfn_name='step_function')
        fact.__sfn_client__ = Mock()
        fact.__sfn_arn__ = "arn"

        test_config_1 = Config(
            workspace="test",
            data={}
        )

        test_config_2 = Config(
            workspace="test",
            data={"key": "value"}
        )

        biz = fact.create('every1', Every(5, Every.MINUTES), configs=[test_config_1])
        biz({'detail-type': 'Scheduled Event',
             'resources': ['arn:aws:events:eu-west-1:123456789:rule/step_function-every1']}, {})
        fact.sfn_client.start_execution.assert_called_once_with(input='{}', stateMachineArn='arn')

        biz = fact.create('every2', Every(10, Every.HOURS), configs=[test_config_2])
        biz({'detail-type': 'Scheduled Event',
             'resources': ['arn:aws:events:eu-west-1:123456789:rule/step_function-every2']}, {})

        fact.sfn_client.start_execution.assert_called_with(input='{"key": "value"}', stateMachineArn='arn')

        assert len(fact.trigger_sources) == 2
        assert [t['name'] for t in fact.trigger_sources] == ['step_function-every1', 'step_function-every2']
        assert [t['value'] for t in fact.trigger_sources] == ['rate(5 minutes)', 'rate(10 hours)']



import os
from unittest.mock import Mock, MagicMock

import pytest

from coworks.cws.fixture import local_server_factory as factory
from tests.mockup import email_mock, smtp_mock, boto3_mock


@pytest.fixture
def example_dir():
    return os.getenv('EXAMPLE_DIR')


@pytest.fixture
def zipfile_mock(request):
    import zipfile

    file = MagicMock()
    file.__enter__ = Mock(return_value=file)
    file.write = Mock()
    zip_mock = Mock(return_value=file)
    zip_mock.file = file

    save_zip_file = zipfile.ZipFile
    zipfile.ZipFile = zip_mock

    def end():
        zipfile.ZipFile = save_zip_file

    request.addfinalizer(end)
    return zip_mock


@pytest.fixture
def email_mock_fixture():
    yield email_mock


@pytest.fixture
def smtp_mock_fixture():
    yield smtp_mock


@pytest.fixture
def boto3_mock_fixture():
    yield boto3_mock


local_server_factory = factory


def pytest_sessionstart():
    example_dir = os.getenv('EXAMPLE_DIR')
    if not os.path.exists(example_dir):
        raise pytest.UsageError(f"Undefined example folder: {example_dir} (value defined in pytest.ini).")



from coworks import TechMicroService
from coworks.cws.runner import CwsRunner


class SimpleMicroService(TechMicroService):

    def get(self):
        return f"Simple microservice ready.\n"


app = SimpleMicroService()
CwsRunner(app)



import os

from chalice.cli import CLIFactory
from chalice.config import Config
from chalice.constants import DEFAULT_APIGATEWAY_STAGE_NAME, DEFAULT_ENDPOINT_TYPE


class CwsFactory(CLIFactory):
    def __init__(self, app, project_dir, **kwargs):
        self.app = app
        super().__init__(project_dir, **kwargs)

    def load_chalice_app(self, environment_variables=None, **kwargs):
        if environment_variables is not None:
            self._environ.update(environment_variables)
            for key, val in self._environ.items():
                os.environ[key] = val
        return self.app

    def mock_config_obj(self, app):
        default_params = {
            'project_dir': self.project_dir,
            'api_gateway_stage': DEFAULT_APIGATEWAY_STAGE_NAME,
            'api_gateway_endpoint_type': DEFAULT_ENDPOINT_TYPE,
            'autogen_policy': False
        }
        config_from_disk = {
            'version': 1,
            'name': 'app',
            'stages': {
            }
        }
        config = Config(user_provided_params={},
                        config_from_disk=config_from_disk,
                        default_params=default_params)

        config._chalice_app = app
        return config





import pytest
import requests

from coworks.tech.s3 import S3MicroService

env = {
    "AWS_ACCESS_KEY_ID": "access key",
    "AWS_SECRET_ACCESS_KEY": "secret key",
    "AWS_REGION": "region"
}


class S3Test(S3MicroService):

    def __init__(self, client, **kwargs):
        super().__init__(**kwargs)
        self.__s3_client__ = client


@pytest.mark.local
class NotDoneTestS3Class:

    def test_list_buckets(self, local_server_factory, boto3_mock_fixture):
        local_server = local_server_factory(S3Test(boto3_mock_fixture, env=env))

        # get list
        response = local_server.make_call(requests.get, '/bucket')
        assert response.status_code == 200
        assert 'Buckets' in response.json()
        names = [b['Name'] for b in response.json()['Buckets']]
        assert "bucket" in names

    def test_create_delete_bucket(self, local_server_factory, boto3_mock_fixture):
        local_server = local_server_factory(S3Test(boto3_mock_fixture, env=env))

        # existing bucket
        response = local_server.make_call(requests.get, '/bucket/bucket')
        assert response.status_code == 200
        response = local_server.make_call(requests.get, '/bucket/test')
        assert response.status_code == 404

        # creates bucket
        response = local_server.make_call(requests.put, '/bucket/test')
        assert response.status_code == 200
        boto3_mock_fixture.create_bucket.assert_called_once_with(bucket='test',
                                                                 createBucketConfiguration={'LocationConstraint': 'EU'})
        response = local_server.make_call(requests.get, '/bucket/test')
        assert response.status_code == 200
        assert response.json()['Name'] == "test"
        boto3_mock_fixture.list_objects.assert_called_with(bucket='test')

        # cannot recreate
        response = local_server.make_call(requests.put, '/bucket/test')
        assert response.status_code == 400

        # creates key
        response = local_server.make_call(requests.get, '/bucket/test', params={'key': 'key'})
        assert response.status_code == 404
        response = local_server.make_call(requests.put, '/content/test', params={'key': 'key', 'body': "content"})
        assert response.status_code == 200
        boto3_mock_fixture.put_object.assert_called_once_with(bucket='test', key='key', body='content')
        response = local_server.make_call(requests.get, '/bucket/test', params={'key': 'key'})
        assert response.status_code == 200
        response = local_server.make_call(requests.get, '/content/test', params={'key': 'key'})
        assert response.status_code == 200
        assert response.text == "content"

        # delete bucket
        response = local_server.make_call(requests.delete, '/bucket/test')
        assert response.status_code == 200
        response = local_server.make_call(requests.get, '/bucket/test')
        assert response.status_code == 404



import json

import requests

from coworks.tech.csv import CSVMicroService

csv_content = json.dumps([
    {"col1": "value1", "col2": "value2"},
    {"col1": "value3", "col2": "value4"},
])


class TestCSV:

    def test_format(self, local_server_factory):
        local_server = local_server_factory(CSVMicroService())
        response = local_server.make_call(requests.post, '/format',
                                          json={'content': csv_content, 'title': False})
        assert response.status_code == 200
        assert response.text == '"value1","value2"\r\n"value3","value4"\r\n'
        response = local_server.make_call(requests.post, '/format',
                                          json={'content': csv_content})
        assert response.status_code == 200
        assert response.text == '"col1","col2"\r\n"value1","value2"\r\n"value3","value4"\r\n'
        response = local_server.make_call(requests.post, '/format',
                                          json={'content': csv_content, 'remove_rows': [0], 'remove_columns': [1]})
        assert response.status_code == 200
        assert response.text == '"col1"\r\n"value3"\r\n'
        response = local_server.make_call(requests.post, '/format',
                                          json={'content': csv_content, 'title': ['test1', 'test2']})
        assert response.status_code == 200
        assert response.text == '"test1","test2"\r\n"value1","value2"\r\n"value3","value4"\r\n'
        response = local_server.make_call(requests.post, '/format',
                                          json={'content': csv_content, 'delimiter': '\t'})
        assert response.status_code == 200
        assert response.text == '"col1"\t"col2"\r\n"value1"\t"value2"\r\n"value3"\t"value4"\r\n'



import os

import pytest

from coworks import TechMicroService
from .runner import ThreadedLocalServer
from ..config import DEFAULT_PROJECT_DIR, DEFAULT_WORKSPACE


@pytest.fixture
def local_server_factory():
    threaded_server = ThreadedLocalServer()

    def create_server(app: TechMicroService, **kwargs):
        project_dir = kwargs.setdefault('project_dir', DEFAULT_PROJECT_DIR)
        workspace = kwargs.setdefault('workspace', DEFAULT_WORKSPACE)

        # if config_path defined, use it to update environment from conf json file
        os.environ['WORKSPACE'] = workspace
        app.deferred_init(workspace)
        app.config.load_environment_variables(project_dir)

        threaded_server.configure(app, **kwargs)
        threaded_server.start()
        return threaded_server

    try:
        yield create_server
    finally:
        threaded_server.shutdown()



import inspect

from coworks import Blueprint, jsonify


class Admin(Blueprint):

    def get_routes(self, pretty=False):
        """Returns the list of entrypoints with signature."""
        app = self._current_app
        routes = {}
        for resource_path, entry in app.routes.items():
            route = {}
            for http_method, route_entry in entry.items():
                function_called = route_entry.view_function
                doc = inspect.getdoc(function_called)
                route[http_method] = {
                    'doc': doc if doc else '',
                    'signature': get_signature(function_called.__cws_func__)
                }
            routes[resource_path] = route

        return jsonify(routes, pretty)

    def get_context(self):
        """Returns the calling context."""
        return self.current_request.to_dict()


def get_signature(func):
    sig = ""
    params = inspect.signature(func).parameters
    index = 0
    for k, p in params.items():
        index += 1
        if index == 1:
            continue
        sp = k
        if p.annotation != inspect.Parameter.empty:
            sp = f"{sp}:{str(p.annotation)}"
        if p.default != inspect.Parameter.empty:
            sp = f"{sp}={p.default}"
        sig = f"{sp}" if index == 2 else f"{sig}, {sp}"
    return f"({sig})"



import sys

from coworks.cws.client import main

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(f"Client test script: Missing number argument")
        sys.exit()

    if sys.argv[1] == "0":
        sys.argv[1:] = ['--project-dir', 'example', 'info', '-m', 'quickstart1']
    elif sys.argv[1] == "1":
        sys.argv[1:] = ['--project-dir', 'example', 'run', '-m', 'quickstart1']
    elif sys.argv[1] == "2":
        sys.argv[1:] = ['--project-dir', 'example', 'export', '-m', 'quickstart2', '-o', 'quickstart.tf']
    elif sys.argv[1] == "3":
        sys.argv[1:] = ['--project-dir', '~/workspace/yopp/yopp_microservices/billing/src', 'export', '-m', 'billing', '-s', 'yopp_orders', '-o', '_billing_yopp_orders.tf']

    # elif sys.argv[1] == "1":
    #     sys.argv[1:] = ['export', '-m', 'example.example', '-s', 'app']
    #
    # elif sys.argv[1] == "2":
    #     sys.argv[1:] = ['export', '-m', 'biz.example', '-s', 'biz', '--format', 'sfn']
    # elif sys.argv[1] == "3":
    #     sys.argv[1:] = ['export', '-m', 'biz.example', '-s', 'biz', '--format', 'sfn', '-b', 'small']
    # elif sys.argv[1] == "4":
    #     sys.argv[1:] = ['--project-dir', 'biz', 'update', '-m', 'example', '-s', 'biz', '-b', 'complete',
    #                     '--profile', 'fpr-customer']

    else:
        print(f"Undefined argument {sys.argv[1]}")
        sys.exit(1)
    main()



import json
import logging
import os
import sys
from collections import defaultdict
from threading import Lock
from typing import Dict, List, Union

from aws_xray_sdk.core import xray_recorder
from chalice import BadRequestError, Rate, Cron
from chalice import Chalice, Blueprint as ChaliceBlueprint
from requests_toolbelt.multipart import MultipartEncoder

from .config import Config
from .mixins import Entry, CoworksMixin, AwsSFNSession
from .utils import class_auth_methods, make_absolute


class Blueprint(CoworksMixin, ChaliceBlueprint):
    """ Represents a blueprint, list of routes that will be added to microservice when registered.

    See :ref:`Blueprint <blueprint>` for more information.

    """

    def __init__(self, name=None, **kwargs):
        """Initialize a blueprint.

        :param kwargs: Other Chalice parameters.

        """
        import_name = name or self.__class__.__name__.lower()
        super().__init__(import_name)

    @property
    def name(self):
        return self._import_name

    @property
    def current_app(self):
        return self._current_app

    @property
    def logger(self):
        return self._current_app.logger

    def entry(self, route):
        return self.current_app.entry(route)

    def deferred_init(self, workspace):
        for func in self.before_first_activation_funcs:
            self.current_app.before_first_activation(func)
        for func in self.before_activation_funcs:
            self.current_app.before_activation(func)
        for func in self.after_activation_funcs:
            self.current_app.after_activation(func)
        for func in self.handle_exception_funcs:
            self.current_app.handle_exception_funcs(func)


class TechMicroService(CoworksMixin, Chalice):
    """Simple tech microservice created directly from class.
    
    See :ref:`tech` for more information.
    
    """

    def __init__(self, name: str = None, configs: Union[Config, List[Config]] = None, **kwargs):
        """ Initialize a technical microservice.
        :param name: Name used to identify the microservice.
        :param configs: Deployment configuration.
        :param workspace used for execution.
        :param kwargs: Other Chalice parameters.
        """
        name = name or self.__class__.__name__.lower()

        self.logger = self.create_logger(name, **kwargs)

        self.configs = configs or [Config()]
        if type(self.configs) is not list:
            self.configs = [configs]
        self.config = None

        super().__init__(app_name=name, **kwargs)

        # Blueprints and extended commands added
        self.experimental_feature_flags.update([
            'BLUEPRINTS'
        ])
        self.blueprints = {}
        self.commands = {}

        # App init deferered functions.
        self.deferred_inits = []
        self._got_first_activation = False
        self._before_activation_lock = Lock()

        self.__auth__ = None
        self.entries = None
        self.sfn_call = False

    @property
    def name(self):
        return self.app_name

    @property
    def ms_type(self):
        return 'tech'

    @staticmethod
    def create_logger(name, **kwargs):
        logger = logging.getLogger(name)
        if 'debug' in kwargs:
            logger.setLevel(logging.INFO)
        return logger

    def deferred_init(self, workspace):
        if self.entries is None:
            self.entries = defaultdict(Entry)

            # Set workspace config
            for conf in self.configs:
                if conf.workspace == workspace:
                    self.config = conf
                    break
            if self.config is None:
                self.config = self.configs[0]

            # Initializes routes
            if self.config.auth:
                self.__auth__ = self._create_auth_proxy(self.config.auth)
            else:
                auth_fun = class_auth_methods(self)
                if auth_fun:
                    self.__auth__ = self._create_auth_proxy(auth_fun)
            self._init_routes(self, authorizer=self.__auth__)

            for deferred_init in self.deferred_inits:
                deferred_init(workspace)
            for blueprint in self.iter_blueprints():
                blueprint.deferred_init(workspace)

    def register_blueprint(self, blueprint: Blueprint, url_prefix=None, authorizer=None, hide_routes=False):
        """ Register a :class:`Blueprint` on the microservice.

        :param blueprint: blueprint to register.
        :param url_prefix: url prefix for the routes added (must be unique for each blueprints).
        :param authorizer: authorizer for this blueprint.
        :param hide_routes:
        :return:
        """
        if url_prefix in self.blueprints:
            raise NameError(f"A blueprint is already defined for the {url_prefix} prefix.")
        self.blueprints[url_prefix] = blueprint

        def deferred(workspace):
            auth = self._create_auth_proxy(authorizer) if authorizer else self.__auth__
            blueprint._init_routes(self, url_prefix=url_prefix, authorizer=auth, hide_routes=hide_routes)
            super(TechMicroService, self).register_blueprint(blueprint, url_prefix=url_prefix)

        self.deferred_inits.append(deferred)

    def entry(self, route):
        route = make_absolute(route)
        route_pathes = route.split('/')
        for entry, result in self.entries.items():
            entry_pathes = entry.split('/')
            if len(route_pathes) != len(entry_pathes):
                continue

            found = True
            for index, path in enumerate(entry_pathes):
                if index < len(route_pathes):
                    if path.startswith('{') or path == route_pathes[index]:
                        continue
                found = False
                break

            if found:
                return result
        return None

    def iter_blueprints(self):
        return self.blueprints.values()

    def execute(self, command, *, project_dir, module=None, service=None, workspace, output=None, error=None,
                **options):
        """Executes a coworks command."""
        from .cws.error import CwsCommandError
        from .cws.client import ProjectConfig  # only available from client module

        module = __name__ if module is None else module
        service = self.name if service is None else service

        project_config = ProjectConfig(project_dir)
        if not project_config.params:
            self.logger.debug(f"No project parameters defined (check {project_config.project_file} file exists).")
        service_config = project_config.get_service_config(module, service, workspace)
        cmd = service_config.get_command(command, self)
        if not cmd:
            raise CwsCommandError(f"The command {command} was not added to the microservice {self.name}.\n")
        command_options = service_config.get_command_options(command)
        execution_params = {**command_options, **options}
        cmd.execute(output=output, error=error, **execution_params)

    def __call__(self, event, context):
        """Lambda handler."""
        try:
            self.do_before_first_activation(event, context)
            self.do_before_activation(event, context)
            response = self.handler(event, context)
            return self.do_after_activation(response)
        except Exception as e:
            print(f"exception: {e}")
            response = self.do_handle_exception(e)
            if response is None:
                raise
            return response

    def handler(self, event, context):
        """Main microservice entry point."""

        # authorization call
        if event.get('type') == 'TOKEN':
            self.logger.debug(f"Calling {self.name} for authorization")

            route = event.get('methodArn').split('/', 3)[-1]
            authorizer = self.entry(f'/{route}').authorizer
            if authorizer:
                return authorizer(event, context)

            self.logger.debug(f"Undefined authorization method for {self.name} ")
            return 'Unauthorized', 403

        # step function call
        if event.get('type') == 'CWS_SFN':
            if self.debug:
                self.logger.debug(f"Calling {self.name} by step function")

            self.sfn_call = True
            content_type = event['headers']['Content-Type']
            if content_type == 'application/json':
                body = event.get('body')
                event['body'] = json.dumps(self._get_data_on_s3(body)) if body else body
            elif content_type == 'multipart/form-data':
                if event.get('form-data'):
                    multi_parts = MultipartEncoder(self._set_multipart_content(event.get('form-data')))
                    event['headers']['Content-Type'] = multi_parts.content_type
                    event['body'] = multi_parts.to_string()
            else:
                raise BadRequestError(f"Undefined content type {content_type} for Step Function call")

        self.logger.debug(f"Calling {self.name} with event {event}")
        res = super().__call__(event, context)

        if self.sfn_call:
            if res['statusCode'] < 200 or res['statusCode'] >= 300:
                raise BadRequestError(f"Status code is {res['statusCode']} : {res['body']}")
            try:
                res['body'] = self._set_data_on_s3(json.loads(res['body']))
            except json.JSONDecodeError:
                pass

        self.logger.debug(f"Call {self.name} returns {res}")
        return res

    def deferred(self, f):
        """Registers a function to be run once the microservice will be initialized.

        May be used as a decorator.

        The function will be called with one parameter, the workspace, and its return value is ignored.
        """

        self.deferred_inits.append(f)
        return f

    def do_before_first_activation(self, event, context):
        """Calls all before first activation functions."""
        if self._got_first_activation:
            return

        # lock needed only if boolean may change value
        with self._before_activation_lock:
            workspace = os.environ['WORKSPACE']
            self.deferred_init(workspace=workspace)

            if not self._got_first_activation:
                for func in self.before_first_activation_funcs:
                    func(event, context)
                self._got_first_activation = True

    def do_before_activation(self, event, context):
        """Calls all before activation functions."""
        for func in self.before_activation_funcs:
            func(event, context)

    def do_after_activation(self, response):
        """Calls all after activation functions."""
        for func in reversed(self.after_activation_funcs):
            resp = func(response)
            if resp is not None:
                response = resp
        return response

    def do_handle_exception(self, exc):
        """Calls all exception handlers."""
        for func in reversed(self.handle_exception_funcs):
            resp = func(exc)
            if resp is not None:
                return resp


class BizFactory(TechMicroService):
    """Tech microservice to create, update and trigger biz microservices.
    """

    def __init__(self, sfn_name, **kwargs):
        super().__init__(name=sfn_name, **kwargs)

        self.aws_profile = self.__sfn_client__ = self.__sfn_arn__ = None
        self.sfn_name = sfn_name
        self.biz: Dict[str, BizMicroService] = {}

        @self.before_first_activation
        def check_sfn():
            return self.sfn_arn

    @property
    def sfn_client(self):
        if self.__sfn_client__ is None:
            session = AwsSFNSession(profile_name=self.aws_profile, env_var_access_key="AWS_RUN_ACCESS_KEY_ID",
                                    env_var_secret_key="AWS_RUN_SECRET_KEY", env_var_region="AWS_RUN_REGION")
            self.__sfn_client__ = session.client
        return self.__sfn_client__

    @property
    def sfn_arn(self):
        if self.__sfn_arn__ is None:
            res = self.sfn_client.list_state_machines()
            while True:
                for sfn in res['stateMachines']:
                    if sfn['name'] == os.environ['SFN_NAME']:
                        self.__sfn_arn__ = sfn['stateMachineArn']
                        return self.__sfn_arn__

                next_token = res.get('nextToken')
                if next_token is None:
                    raise BadRequestError(f"Undefined step function : {self.sfn_name}")

                res = self.sfn_client.list_state_machines(nextToken=next_token)
        return self.__sfn_arn__

    @property
    def trigger_sources(self):

        def to_dict(name, trigger):
            return {'name': f"{self.sfn_name}-{name}", **trigger.to_dict()}

        return [to_dict(name, biz.trigger) for name, biz in self.biz.items()]

    def create(self, biz_name, trigger=None, configs: List[Config] = None, **kwargs):
        """Creates a biz microservice. If the trigger is not defined the microservice can only be triggered manually."""

        if biz_name in self.biz:
            raise BadRequestError(f"Biz microservice {biz_name} already defined for {self.sfn_name}")

        self.biz[biz_name] = BizMicroService(self, trigger, configs, name=biz_name, **kwargs)
        return self.biz[biz_name]

    def invoke(self, data):
        res = self.sfn_client.start_execution(stateMachineArn=self.sfn_arn, input=json.dumps(data if data else {}))
        return json.dumps(res, indent=4, sort_keys=True, default=str)


class BizMicroService(TechMicroService):
    """Biz composed microservice activated by a reactor.
    """

    def __init__(self, biz_factory, trigger, configs, **kwargs):
        super().__init__(**kwargs)
        self.biz_factory = biz_factory
        self.configs = configs
        self.trigger = trigger

    @property
    def ms_type(self):
        return 'biz'

    @property
    def trigger_source(self):
        return self.trigger.to_dict()

    def get_sfn_name(self):
        """Returns the name of the associated step function."""
        return self.biz_factory.sfn_name

    def get_sfn_arn(self):
        """Returns the arn of the associated step function."""
        return self.biz_factory.sfn_arn

    def get_biz_names(self):
        """Returns the list of biz microservices defined in the factory."""
        return [name for name in self.biz_factory]

    def get_default_data(self):
        try:
            workspace = os.environ['WORKSPACE']
        except KeyError:
            raise EnvironmentError("environment variable WORKSPACE is not defined")
        try:
            default_data = next(c.data for c in self.configs if c.workspace == workspace)
        except StopIteration:
            self.logger.debug(f"No configuration found for workspace {workspace} in {self.configs}")
            default_data = {}
        except KeyError:
            default_data = {}
        return default_data

    def post_trigger(self, data=None):
        data = data or {}
        data.update(self.get_default_data())
        return self.biz_factory.invoke(data)

    def handler(self, event, context):
        if 'detail-type' in event and event['detail-type'] == 'Scheduled Event':
            if self.debug:
                self.logger.debug(f"Trigger: {self.biz_factory.sfn_name}")
            return self.biz_factory.invoke(self.get_default_data())

        return super().handler(event, context)


def hide(f):
    """Hide a route of the microservice.

     May be used as a decorator.

     Usefull when creating inherited microservice.
     """

    setattr(f, '__cws_hidden', True)
    return f


class Once:
    ...


class At(Cron):
    def __init__(self, minutes=0, hours=1, day_of_month=None, month=None, day_of_week=None, year=None):
        super().__init__(minutes, hours, day_of_month, month, day_of_week, year)

    def to_dict(self):
        return {
            'source': 'at',
            'value': self.to_string(),
        }


class Every(Rate):
    def __init__(self, value, unit):
        super().__init__(value, unit)

    def to_dict(self):
        return {
            'source': 'every',
            'value': self.to_string(),
        }



import os
import time
from http.client import BadStatusLine
from pyexpat import ExpatError
from typing import List, Tuple, Union
from xmlrpc import client
from xmlrpc.client import Fault, ProtocolError

from aws_xray_sdk.core import xray_recorder

from coworks.mixins import Boto3Mixin
from .. import Blueprint


class Odoo(Blueprint, Boto3Mixin):
    def __init__(self, url_env_var_name='URL', dbname_env_var_name='DBNAME', user_env_var_name='USER', passwd_env_var_name='PASSWD', **kwargs):
        super().__init__(**kwargs)
        self.url = self.dbname = self.user = self.passwd = None
        self.url_env_var_name = url_env_var_name
        self.dbname_env_var_name = dbname_env_var_name
        self.user_env_var_name = user_env_var_name
        self.passwd_env_var_name = passwd_env_var_name
        self.api_uid = None

    def get(self):
        """Check the connection."""
        self.connect()
        return "connected"

    def get_model(self, model: str, searched_field: str, searched_value, fields=None, ensure_one=False, **kwargs):
        """Returns the list of objects or the object which searched_field is equal to the searched_value."""
        fields = fields or ['id']
        results = self.search(model, [[(searched_field, '=', searched_value)]], fields=fields, **kwargs)
        return results

    def get_fields(self, model, searched_field, searched_value, returned_fields=None):
        """Returns the value of the object which searched_field is equal to the searched_value."""
        returned_fields = returned_fields or ['id']
        return self.get_model(model, searched_field, searched_value, fields=returned_fields, ensure_one=True)

    def get_field(self, model, searched_field, searched_value, returned_field='id'):
        """Returns the value of the object which searched_field is equal to the searched_value."""
        return self.get_model(model, searched_field, searched_value, fields=[returned_field], ensure_one=True)[0]

    def get_id(self, model, searched_field, searched_value):
        """Returns the id of the object which searched_field is equal to the searched_value."""
        return self.get_field(model, searched_field, searched_value)

    @xray_recorder.capture("Connect to ODOO")
    def connect(self, url=None, dbname=None, user=None, passwd=None):
        # initialize connection informations
        self.url = url or os.getenv(self.url_env_var_name)
        if not self.url:
            raise EnvironmentError(f"{self.url_env_var_name} not defined in environment.")
        self.dbname = dbname or os.getenv(self.dbname_env_var_name)
        if not self.dbname:
            raise EnvironmentError(f"{self.dbname_env_var_name} not defined in environment.")
        self.user = user or os.getenv(self.user_env_var_name)
        if not self.user:
            raise EnvironmentError(f"{self.user_env_var_name} not defined in environment.")
        self.passwd = passwd or os.getenv(self.passwd_env_var_name)
        if not self.passwd:
            raise EnvironmentError(f"{self.passwd_env_var_name} not defined in environment.")

        try:
            subsegment = xray_recorder.current_subsegment()
            if subsegment:
                subsegment.put_metadata("connection", f"Connection parameters: {self.url}, {self.dbname}, {self.user}, {self.passwd}")

            # initialize xml connection to odoo
            common = client.ServerProxy(f'{self.url}/xmlrpc/2/common')
            self.api_uid = common.authenticate(self.dbname, self.user, self.passwd, {})
            if not self.api_uid:
                raise Exception(f'Odoo connection parameters are wrong.')
        except Exception:
            raise Exception(f'Odoo interface variables wrongly defined.')

    @xray_recorder.capture()
    def execute_kw(self, model: str, method: str, *args):
        if not self.api_uid:
            self.connect()
        models_url = f'{self.url}/xmlrpc/2/object'
        try:
            with client.ServerProxy(models_url, allow_none=True) as models:
                res: List[dict] = models.execute_kw(self.dbname, self.api_uid, self.passwd, model, method, *args)
                return res
        except (BadStatusLine, ExpatError, ProtocolError):
            time.sleep(2)
            with client.ServerProxy(models_url) as models:
                return models.execute_kw(self.dbname, self.api_uid, self.passwd, model, method, *args)
        except Fault as e:
            if e.faultString.endswith("TypeError: cannot marshal None unless allow_none is enabled\n"):
                return
            raise
        except Exception as e:
            raise

    def search(self, model, filters: List[List[tuple]], *, fields=None, offset=None, limit=None, order=None,
               ensure_one: bool = False) -> Union[Tuple[str, int], dict, List[dict]]:
        options = {}
        if fields:
            options["fields"] = fields if type(fields) is list else [fields]
        options.setdefault("offset", int(offset) if offset else 0)
        options.setdefault("limit", int(limit) if limit else 50)
        options.setdefault("order", order if order else 'id asc')

        try:
            results = self.execute_kw(model, 'search_read', filters, options)
            return _reduce_result(results, fields, ensure_one)
        except Exception as e:
            return str(e), 404

    def create(self, model, data: dict):
        return self.execute_kw(model, 'create', [self._replace_tuple(data)])

    def write(self, model, _id, data: dict):
        return self.execute_kw(model, 'write', [[_id], self._replace_tuple(data)])

    # def delete(self, model, _id):
    #     return self.execute_kw(model, 'unlink', [[_id]])

    def _replace_tuple(self, struct: dict) -> dict:
        """For data from JSON, tuple are defined with key surronded by paranthesis."""
        for k, value in struct.items():
            if isinstance(value, dict):
                self._replace_tuple(value)
            else:
                if k.startswith('(') and type(value) is list:
                    del struct[k]
                    struct[k[1:-1]] = [tuple(v) for v in value]
        return struct


def _reduce_result(results: List[dict], fields, ensure_one):
    """Reduces rows values if only one value and ensure only only one in the result list and returns it."""
    if len(fields) == 1:
        results = [row[fields[0]] for row in results]
    if ensure_one:
        if len(results) == 0:
            return "No object found.", 404
        if len(results) > 1:
            return f"More than one object ({len(results)}) founds : ids={[o.get('id') for o in results]}", 404
        return results[0]
    return results



from unittest.mock import Mock

import click
import pytest

from coworks.cws.command import CwsCommand, CwsCommandError
from tests.src.coworks.tech_ms import *

cmd_mock = Mock()


class MyCommand(CwsCommand):
    def _execute(self, *, project_dir, **options):
        cmd_mock()
        assert project_dir == 'tests/src/cws'
        assert options['module'] == 'test_cmd'
        assert options['workspace'] == 'dev'
        assert options['service'] == 'test'


class MyCommandWithOptions(MyCommand):

    @property
    def options(self):
        return [
            *super().options,
            click.option('--param', required=True),
            click.option('--autre', required=True),
        ]


class TestCommand:

    def test_command(self):
        simple = SimpleMS()
        MyCommand(simple, name='test')

        with pytest.raises(CwsCommandError) as pytest_wrapped_e:
            simple.execute('autre', project_dir='tests/src/cws', module='test_cmd', workspace='dev')
        assert pytest_wrapped_e.type == CwsCommandError
        assert pytest_wrapped_e.value.msg == 'The command autre was not added to the microservice test.\n'

        simple.execute('test', project_dir='tests/src/cws', module='test_cmd', workspace='dev')
        cmd_mock.assert_called_once()

        simple.execute('test', project_dir='tests/src/cws', module='test_cmd', workspace='dev', help=None)

    def test_command_with_options(self):
        simple = SimpleMS()
        MyCommandWithOptions(simple, name='test_command_with_options')

        with pytest.raises(CwsCommandError) as pytest_wrapped_e:
            simple.execute('test_command_with_options', project_dir='tests/src/cws', module='test_cmd', workspace='dev')
        assert pytest_wrapped_e.type == CwsCommandError
        assert pytest_wrapped_e.value.msg == 'missing parameter: param'

        simple.execute('test_command_with_options', project_dir='tests/src/cws', module='test_cmd', workspace='dev',
                       param='param')



from unittest.mock import Mock

import requests

from coworks.middleware.xray_middleware import XRayMiddleware
from tests.src.coworks.tech_ms import *


def const_id(*args):
    return lambda x: x


xray_recorder = Mock()
segment = Mock()
xray_recorder.capture = Mock(side_effect=const_id)
xray_recorder.current_subsegment = Mock(return_value=segment)


class TestMiddleware:
    def __init__(self, app):
        app.before_first_activation(self.always_first)
        app.before_activation(self.before)
        app.after_activation(self.at_least)
        app.handle_exception(self.on_exception)
        self.evts = []

    def always_first(self, event, context):
        self.evts.append("first")

    def before(self, event, context):
        self.evts.append("before")

    def at_least(self, response):
        if response['statusCode'] == 400:
            self.evts.append("error")
        else:
            self.evts.append("after")
        return response

    def on_exception(self, exc):
        self.evts.append("exception")


class TestClass:

    def test_decorators(self, local_server_factory):
        simple = SimpleMS()
        evts = []

        @simple.before_first_activation
        def always_first(event, context):
            evts.append("first")

        @simple.before_activation
        def before(event, context):
            evts.append("before")

        @simple.after_activation
        def at_least(response):
            if response['statusCode'] == 400:
                evts.append("error")
            else:
                evts.append("after")
            return response

        @simple.handle_exception
        def on_exception(exc):
            evts.append("exception")

        local_server = local_server_factory(simple)
        local_server.make_call(requests.get, '/')
        local_server.make_call(requests.get, '/')
        assert evts == ["first", "before", "after", "before", "after"]

    def test_middleware(self, local_server_factory):
        simple = SimpleMS()
        middleware = TestMiddleware(simple)
        local_server = local_server_factory(simple)
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == 'get'
        assert middleware.evts == ["first", "before", "after"]
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert middleware.evts == ["first", "before", "after", "before", "after"]
        response = local_server.make_call(requests.get, '/?wrong=1')
        assert response.status_code == 400
        assert 'error' in middleware.evts

    def test_xray_middleware(self, local_server_factory):
        simple = SimpleMS()
        XRayMiddleware(simple, xray_recorder)
        local_server = local_server_factory(simple)
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == 'get'
        xray_recorder.capture.assert_called()
        segment.put_annotation.assert_called()



import json
import os
import re
from dataclasses import dataclass
from json import JSONDecodeError
from pathlib import Path
from typing import Callable, Union, List, Tuple, Any

from chalice import CORSConfig as ChaliceCORSConfig
from chalice.app import AuthRequest as ChaliceAuthRequest

from .mixins import CoworksMixin
from .utils import as_list

DEFAULT_PROJECT_DIR = '.'
DEFAULT_WORKSPACE = 'dev'

ENV_FILE_SUFFIX = '.json'
SECRET_ENV_FILE_SUFFIX = '.secret.json'


class AuthRequest(ChaliceAuthRequest):
    def __init__(self, auth_type, token, method_arn):
        self.auth_type = auth_type
        self.token = token
        self.method_arn = method_arn

        _, self.workspace, self.method, self.route = method_arn.split('/', 3)


class CORSConfig(ChaliceCORSConfig):

    def get_access_control_headers(self):
        if not self.allow_origin:
            return {}
        return super().get_access_control_headers()


@dataclass
class Config:
    """ Configuration class for deployment."""

    version: str = ""
    workspace: str = DEFAULT_WORKSPACE
    environment_variables_file: Union[str, List[str]] = None
    environment_variables: Union[dict, List[dict]] = None
    auth: Callable[[CoworksMixin, AuthRequest], Any] = None
    cors: CORSConfig = CORSConfig(allow_origin='')
    content_type: Tuple[str] = ('multipart/form-data', 'application/json', 'text/plain')
    data: dict = None

    def existing_environment_variables_files(self, project_dir):
        """Returns a list containing the paths to environment variables files that actually exist """
        environment_variables_file = as_list(self.environment_variables_file)
        files = []
        parent = Path(project_dir)
        for file in environment_variables_file:
            var_file = parent / file
            if var_file.is_file():
                files.append(var_file)
            var_secret_file = var_file.with_suffix(SECRET_ENV_FILE_SUFFIX)
            if var_secret_file.is_file():
                files.append(var_secret_file)
        return files

    def load_environment_variables(self, project_dir):
        """Uploads environment variables from the environment variables files and variables."""
        environment_variables = {}

        # Environment variables from files and from config
        for file in self.existing_environment_variables_files(project_dir):
            environment_variables.update(self._load_file(file))
        if self.environment_variables:
            environment_variables.update(self.environment_variables)

        # Check environment variables name are accepted by AWS
        # Keys start with a letter and are at least two characters.
        # Keys only contain letters, numbers, and the underscore character (_).
        var_name_regexp = re.compile(r'[a-zA-Z][a-zA-Z_]+')
        for key in environment_variables:
            if not var_name_regexp.fullmatch(key):
                raise KeyError(f'Wrong environment variable name: {key}')

        # Set environment variables
        if environment_variables:
            for key, value in environment_variables.items():
                os.environ[key] = str(value)

    def setdefault(self, key, value):
        """Same as for dict."""
        if not hasattr(self, key):
            setattr(self, key, value)

    @staticmethod
    def _load_file(var_file):
        try:
            with var_file.open() as f:
                return json.loads(f.read())
        except JSONDecodeError as e:
            raise FileNotFoundError(f"Syntax error when in e{var_file}: {str(e)}.\n")
        except Exception as e:
            print(type(e))
            raise FileNotFoundError(f"Error when loading environment variables files {str(e)}.\n")



import threading
import time
from pathlib import Path

import pytest
import requests

from coworks.config import Config
from coworks.cws.client import client
from coworks.cws.runner import ThreadedLocalServer
from coworks.utils import import_attr
from .example import TechMS


class TestClass:
    def test_simple_example(self, local_server_factory):
        local_server = local_server_factory(TechMS())
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == "Simple microservice for test.\n"
        response = local_server.make_call(requests.get, '/', params={"usage": "demo"})
        assert response.status_code == 200
        assert response.text == "Simple microservice for demo.\n"
        response = local_server.make_call(requests.get, '/value/1')
        assert response.status_code == 200
        assert response.text == "0\n"
        response = local_server.make_call(requests.put, '/value/1', json={"value": 456})
        assert response.status_code == 200
        assert response.text == "456"
        response = local_server.make_call(requests.get, '/value/1')
        assert response.status_code == 200
        assert response.text == "456\n"

    def test_params(self, local_server_factory):
        local_server = local_server_factory(TechMS())
        response = local_server.make_call(requests.put, '/value/1', json=456)
        assert response.status_code == 200
        assert response.text == "456"
        response = local_server.make_call(requests.get, '/value/1')
        assert response.status_code == 200
        assert response.text == "456\n"

    def test_env(self, local_server_factory, example_dir):
        config = Config(environment_variables_file=Path(example_dir) / "config" / "vars_dev.json")
        local_server = local_server_factory(TechMS(configs=[config]))
        response = local_server.make_call(requests.get, '/env', timeout=500)
        assert response.status_code == 200
        assert response.text == "Simple microservice for test dev environment variable.\n"

    def test_init(self, local_server_factory, example_dir):
        local_server = local_server_factory(TechMS(configs=[]))
        response = local_server.make_call(requests.get, '/init', timeout=500)
        assert response.status_code == 200
        assert response.text == "Initial value is test.\n"

    def test_run_example(self, example_dir):
        app = import_attr('example', 'app', cwd=example_dir)
        port = ThreadedLocalServer.unused_tcp_port()
        server = threading.Thread(target=run_server_example, args=(app, port, example_dir), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{port}/')
        assert response.text == "Simple microservice for test.\n"
        response = requests.get(f'http://localhost:{port}/', params={"usage": "demo"})
        assert response.status_code == 200
        assert response.text == "Simple microservice for demo.\n"

    def test_cmd_wront_project_dir(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests/example', 'info'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == \
               'Microservice project1 defined in module example\nMicroservice project2 defined in module example\n'


def run_server_example(app, port, example_dir):
    print(f"Server starting on port {port}")
    app.execute('run', project_dir=example_dir, module='example', workspace='dev', host='localhost', port=port, )



import os
from collections import defaultdict

import click

from coworks import TechMicroService
from coworks.config import Config
from coworks.cws.command import CwsCommand
from coworks.cws.runner import CwsRunner
from coworks.cws.writer import CwsTerraformWriter


class TestCmd(CwsCommand):

    @property
    def options(self):
        return [
            *super().options,
            click.option('-a', '--a', required=True),
            click.option('--b')
        ]

    def _execute(self, *, a, b, **options):
        if a:
            self.output.write(f"test command with a={a}")
            self.output.write("/")
        self.output.write(f"test command with b={b}")


class TechMS(TechMicroService):
    """Technical microservice for the Coworks tutorial example."""
    version = "1.2"
    values = defaultdict(int)
    init_value = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        @self.deferred
        def init(workspace):
            self.init_value = 'test'

    def get(self, usage="test"):
        """Entrypoint for testing named parameter."""
        return f"Simple microservice for {usage}.\n"

    def get_value(self, index):
        """Entrypoint for testing positional parameter."""
        return f"{self.values[index]}\n"

    def put_value(self, index, value=0):
        self.values[index] = value
        return value

    def get_init(self):
        return f"Initial value is {self.init_value}.\n"

    def get_env(self):
        return f"Simple microservice for {os.getenv('test')}.\n"


# usefull for test info (don't remove)
tech_app = TechMS()
TestCmd(tech_app, name='test')
CwsTerraformWriter(tech_app)

app = TechMS(configs=Config(environment_variables_file="config/vars_dev.json"))
TestCmd(tech_app, name='test')


class RunnerMock(CwsRunner):

    def _execute(self, *, project_dir, workspace, host, port, debug, **options):
        assert type(port) is int


project1 = TechMS()
RunnerMock(project1)
project2 = TechMS()



import requests

from coworks.config import Config, CORSConfig
from .tech_ms import *


class NoneCorsMS(TechMicroService):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self):
        """Root access."""
        return "get"


class AllCorsMS(TechMicroService):

    def __init__(self, **kwargs):
        config = Config(cors=CORSConfig(allow_origin='*'))
        super().__init__(configs=config, **kwargs)

    def get(self):
        """Root access."""
        return "get"


class OneCorsMS(TechMicroService):

    def __init__(self, **kwargs):
        config = Config(cors=CORSConfig(allow_origin='www.test.fr'))
        super().__init__(configs=config, **kwargs)

    def get(self):
        """Root access."""
        return "get"


class SeveralCorsMS(TechMicroService):

    def __init__(self, **kwargs):
        config = Config(cors=CORSConfig(allow_origin=['www.test.fr', 'www.test.com']))
        super().__init__(configs=config, **kwargs)

    def get(self):
        """Root access."""
        return "get"


class OtherCorsMS(TechMicroService):

    def __init__(self, **kwargs):
        config = Config(cors=CORSConfig(allow_origin='https://www.test.fr',
                                        allow_headers=['X-Special-Header'],
                                        max_age=600,
                                        expose_headers=['X-Special-Header'],
                                        allow_credentials=True))
        super().__init__(configs=config, **kwargs)

    def get(self):
        """Root access."""
        return "get"


class TestClass:
    def test_authorize_none(self, local_server_factory):
        ms = NoneCorsMS()
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.options, '/', timeout=500)
        assert response.status_code == 200
        assert 'Access-Control-Allow-Origin' not in response.headers

    def test_authorize_all(self, local_server_factory):
        ms = AllCorsMS()
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.options, '/')
        assert response.status_code == 200
        assert response.headers['Access-Control-Allow-Origin'] == '*'

    def test_authorize_one(self, local_server_factory):
        ms = OneCorsMS()
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.options, '/')
        assert response.status_code == 200
        assert response.headers['Access-Control-Allow-Origin'] == 'www.test.fr'

    def test_authorize_several(self, local_server_factory):
        ms = SeveralCorsMS()
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.options, '/')
        assert response.status_code == 200
        assert response.headers['Access-Control-Allow-Origin'] == 'www.test.fr, www.test.com'

    def test_authorize_other(self, local_server_factory):
        ms = OtherCorsMS()
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.options, '/')
        assert response.status_code == 200
        assert response.headers['Access-Control-Allow-Origin'] == 'https://www.test.fr'
        assert response.headers['Access-Control-Max-Age'] == '600'



import smtplib
from email import message
from unittest.mock import Mock, MagicMock

#
# Email
#

email_mock = message.EmailMessage()
email_mock.set_content = Mock()
email_mock.add_attachment = Mock()


class EmailMessageMock:
    def __new__(cls, *args, **kwargs):
        return email_mock


message.EmailMessage = EmailMessageMock

smtp_mock = MagicMock()
smtp_mock.__enter__.return_value = smtp_mock
smtp_mock.send_message.return_value = "Mail sent to you@test.com"


class SMTPMock:
    def __new__(cls, *args, **kwargs):
        return smtp_mock


smtplib.SMTP = SMTPMock

#
# S3
#

boto3_mock = MagicMock()


class S3Reader():

    def __init__(self, content):
        self.content = content

    def read(self):
        return self.content.encode()

    def write(self, content):
        self.content = content


class BucketMockup:
    BUCKETS = {'bucket': {'key': S3Reader('content')}}

    @classmethod
    def add_bucket(cls, bucket='', **kwargs):
        cls.BUCKETS[bucket] = {}

    @classmethod
    def get_bucket(cls, bucket=''):
        return cls.BUCKETS.get(bucket)

    @classmethod
    def delete_bucket(cls, bucket=''):
        del cls.BUCKETS[bucket]

    @classmethod
    def get_list(cls):
        return {"Buckets": [{'Name': k} for k in cls.BUCKETS]}

    @classmethod
    def get_object(cls, bucket='', key=''):
        buck = cls.get_bucket(bucket)
        return {"Body": buck.get(key), "Name": key} if key in buck else {}

    @classmethod
    def put_content(cls, bucket='', key='', body=''):
        obj = cls.get_object(bucket)
        if obj:
            obj["Body"].write(body)
        else:
            cls.get_bucket(bucket)[key] = S3Reader(body)


boto3_mock.list_buckets = Mock(side_effect=BucketMockup.get_list)
boto3_mock.list_objects = Mock(side_effect=BucketMockup.get_bucket)

boto3_mock.create_bucket = Mock(side_effect=BucketMockup.add_bucket)
boto3_mock.delete_object.return_value = {"Buckets": [{"Name": 'bucket'}]}
boto3_mock.delete_bucket = Mock(side_effect=BucketMockup.delete_bucket)

boto3_mock.get_object = Mock(side_effect=BucketMockup.get_object)
boto3_mock.put_object = Mock(side_effect=BucketMockup.put_content)
# boto3_mock.delete_object = Mock(side_effect=s3_delete)



import traceback
from functools import partial, update_wrapper

LAMBDA_NAMESPACE = 'lambda'
COWORKS_NAMESPACE = 'coworks'


class XRayMiddleware:

    def __init__(self, app, recorder):
        self.app = app
        self.app.logger.info("initializing xray middleware")

        @app.before_first_activation
        def capture_routes(event, context):
            for path, route in app.routes.items():
                for method, entry in route.items():
                    view_function = entry.view_function
                    cws_function = view_function.__cws_func__

                    def captured(_view_function, *args, **kwargs):
                        subsegment = recorder.current_subsegment()
                        subsegment.put_metadata('event', event, LAMBDA_NAMESPACE)
                        subsegment.put_metadata('context', context, LAMBDA_NAMESPACE)
                        subsegment.put_annotation('service', app.name)
                        subsegment.put_metadata('query_params', app.current_request.query_params, COWORKS_NAMESPACE)
                        if app.current_request.raw_body:
                            subsegment.put_metadata('json_body', app.current_request.json_body, COWORKS_NAMESPACE)
                        response = _view_function(*args, **kwargs)
                        subsegment.put_metadata('response', response, COWORKS_NAMESPACE)
                        return response

                    def captured_entry(_cws_function, *args, **kwargs):
                        subsegment = recorder.current_subsegment()
                        subsegment.put_metadata('args', args, COWORKS_NAMESPACE)
                        subsegment.put_metadata('kwargs', kwargs, COWORKS_NAMESPACE)
                        response = _cws_function(*args, **kwargs)
                        subsegment.put_metadata('response', response, COWORKS_NAMESPACE)
                        return response

                    wrapped_fun = update_wrapper(partial(captured, view_function), view_function)
                    entry.view_function = recorder.capture(view_function.__name__)(wrapped_fun)

                    wrapped_fun = update_wrapper(partial(captured_entry, cws_function), cws_function)
                    entry.view_function.__cws_func__ = recorder.capture(cws_function.__name__)(wrapped_fun)

        @app.handle_exception
        def capture_exception(e):
            app.logger.error(e)
            app.logger.error(traceback.extract_stack())
            subsegment = recorder.current_subsegment
            if subsegment:
                subsegment.put_annotation('service', app.name)
                subsegment.add_exception(e, traceback.extract_stack())

            return 'Exception in microservice', 200



import os
import threading
import time
from pathlib import Path

import pytest
import requests

from coworks.config import Config
from coworks.cws.runner import CwsRunner
from coworks.cws.runner import ThreadedLocalServer
from tests.src.coworks.tech_ms import *

EXAMPLE_DIR = os.getenv('EXAMPLE_DIR')


class WithEnvMS(SimpleMS):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        @self.before_first_activation
        def init(event, context):
            assert os.getenv("test") is not None

    def get(self):
        """Root access."""
        return os.getenv("test")


class TestClass:

    def test_dev_stage(self, local_server_factory):
        config = Config(environment_variables_file=Path(EXAMPLE_DIR) / "config" / "vars_dev.json")
        local_server = local_server_factory(WithEnvMS(configs=config))
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == 'test dev environment variable'

    def test_run_dev_stage(self, example_dir):
        config = Config(environment_variables_file=Path(EXAMPLE_DIR) / "config" / "vars_dev.json")
        app = WithEnvMS(configs=config)
        CwsRunner(app)
        port = ThreadedLocalServer.unused_tcp_port()
        server = threading.Thread(target=run_server_example, args=(app, port), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{port}/', headers={'Authorization': "token"})
        assert response.text == "test dev environment variable"

    def test_prod_stage(self, local_server_factory):
        config = Config(environment_variables_file=Path(EXAMPLE_DIR) / "config" / "vars_prod.json")
        local_server = local_server_factory(WithEnvMS(configs=config))
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == 'test prod environment variable'

    def test_env_var(self, local_server_factory):
        config = Config(environment_variables={'test': 'test value environment variable'})
        local_server = local_server_factory(WithEnvMS(configs=config))
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == 'test value environment variable'

    def test_wrong_env_var_name(self, local_server_factory):
        config = Config(environment_variables={'1test': 'test value environment variable'})
        with pytest.raises(KeyError) as pytest_wrapped_e:
            local_server = local_server_factory(WithEnvMS(configs=config))
        assert pytest_wrapped_e.type == KeyError
        assert pytest_wrapped_e.value.args[0] == "Wrong environment variable name: 1test"


def run_server_example(app, port):
    print(f"Server starting on port {port}")
    app.execute('run', host='localhost', port=port, project_dir='.', module='example', workspace='dev')



from coworks import TechMicroService
from coworks.config import Config
from coworks.cws.runner import CwsRunner
from coworks.cws.writer import CwsTerraformWriter


class SimpleMicroService(TechMicroService):

    def get(self):
        return f"Simple microservice ready.\n"

    def auth(self, auth_request):
        return auth_request.token == "token"


CONFIG = Config(
    workspace="dev"
)

app = SimpleMicroService(name='test', configs=[CONFIG])
CwsRunner(app)
CwsTerraformWriter(app, name='export')

if __name__ == '__main__':
    app.execute('run', project_dir='.', workspace='dev', module='quickstart')



from unittest.mock import MagicMock

from coworks import TechMicroService

client = MagicMock()
session = MagicMock()
session.client = client


class TechMS(TechMicroService):

    def get_test(self):
        return "get"


class LambdaContext:

    def __init__(self):
        self.aws_request_id = "id"
        self.function_name = "fun"


class TestClass:
    def test_save_on_s3(self):
        data = {
            'short': "normal",
            'long': "x" * 10000,
        }
        tech: TechMS = TechMS()
        tech.aws_s3_sfn_data_session = session
        tech.lambda_context = LambdaContext()
        res = tech._set_data_on_s3(data)
        client.put_object.assert_called()
        assert res['short'] == 'normal'
        assert res['long'].startswith('$$')
        assert res['long'].endswith('$$')



import io
import json
import os
from unittest.mock import Mock, MagicMock

import pytest
import yaml

from coworks import BizFactory
from coworks.cws.sfn import StepFunction, TechState, CwsSFNTranslater
from coworks.cws.writer import CwsWriterError
from tests.src.coworks.tech_ms import S3MockTechMS


class TestStepFunction(StepFunction):

    def __init__(self, data):
        filepath = Mock()
        file = MagicMock()
        file.__enter__ = Mock(return_value=data)
        filepath.open = Mock(return_value=file)
        with filepath.open() as file:
            self.data = yaml.load(file, Loader=yaml.SafeLoader)

        options = {'module': None, 'service': None, 'project_dir': None, 'workspace': None}
        super().__init__("test", filepath, account_number='123412341234', **options)


class TechMS(S3MockTechMS):

    def get_test(self):
        return "get"

    def get_params(self, value, other):
        return f"get {value} and {other}"

    def get_params_(self, value=1, other=2):
        return f"get {value} and {other}"

    def post_params(self, value=1, other=2):
        return f"get {value} and {other}"


class TestClass:
    os.environ["WORKSPACE"] = ''  # not sure where this should be set

    def test_no_params(self):
        tech = TechMS()

        data = {'get': '/test'}
        call = TechState.get_call_data(None, data)
        res = tech(call, {})
        assert res['statusCode'] == 200

    def test_arg_params(self):
        tech = TechMS()

        uri_params = {'_0': 1, '_1': 2}
        data = {'get': '/params/{_0}/{_1}', 'uri_params': uri_params}
        call = TechState.get_call_data(None, data)
        res = tech(call, {})
        assert res['statusCode'] == 200
        assert res['body'] == "get 1 and 2"

    def test_kwargs_params(self):
        tech = TechMS()

        data = {'get': '/params'}
        call = TechState.get_call_data(None, data)
        res = tech(call, {})
        assert res['statusCode'] == 200
        assert res['body'] == "get 1 and 2"

        query_params = {'value': [3], 'other': [4]}
        data = {'get': '/params', 'query_params': query_params}
        call = TechState.get_call_data(None, data)
        res = tech(call, {})
        assert res['statusCode'] == 200
        assert res['body'] == "get 3 and 4"

        query_params = {'value': [5], 'other': [6]}
        data = {'post': '/params', 'query_params': query_params}
        call = TechState.get_call_data(None, data)
        res = tech(call, {})
        assert res['statusCode'] == 200
        assert res['body'] == "get 5 and 6"

        body = {'value': 7, 'other': 8}
        data = {'post': '/params', 'body': body}
        call = TechState.get_call_data(None, data)
        res = tech(call, {})
        assert res['statusCode'] == 200
        assert res['body'] == "get 7 and 8"

    def test_biz_empty(self):
        biz = BizFactory('tests/src/coworks/biz/empty')
        biz.create('test')
        translater = CwsSFNTranslater(biz)
        output = io.StringIO()
        with pytest.raises(CwsWriterError):
            options = {'project_dir': '.', 'module': '', 'service': '', 'workspace': 'dev'}
            translater.execute(output=output, error=output, **options)

    def test_biz_complete(self):
        """Tests the doc example."""
        fact = BizFactory('tests/src/coworks/biz/complete')
        fact.create('test')
        translater = CwsSFNTranslater(fact)
        output = io.StringIO()
        options = {'project_dir': '.', 'module': '', 'service': '', 'workspace': 'dev'}
        translater.execute(output=output, error=output, **options)
        output.seek(0)
        source = json.loads(output.read())
        assert source['Version'] == "1.0"
        assert 'Comment' in source
        assert len(source['States']) == 4

        states = source['States']
        state = states['Check server']
        assert state is not None
        assert state['Type'] == 'Task'

        state = states['Send mail']
        assert state is not None
        assert state['Type'] == 'Task'
        assert state['End'] is True

    def test_fail(self):
        data = {'states': [{
            'name': "fail",
            'fail': None
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        sfn.generate()
        assert len(sfn.all_states) == 2
        assert 'End' not in sfn.all_states[1].state
        assert 'Cause' in sfn.all_states[1].state
        assert 'Error' in sfn.all_states[1].state

    def test_pass(self):
        # missing key cases
        data = {'states': [{
            'name': "action",
            'pass': None
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        sfn.generate()
        assert len(sfn.all_states) == 2
        assert 'End' in sfn.all_states[1].state

    def test_tech(self):
        # missing key cases
        data = {'states': [{
            'name': "action",
            'tech': {
                'service': "tech",
            }
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        with pytest.raises(CwsWriterError) as pytest_wrapped_e:
            sfn.generate()
        assert pytest_wrapped_e.type == CwsWriterError
        assert pytest_wrapped_e.value.args[0] == "No route defined for {'name': 'action', 'tech': {'service': 'tech'}}"

        data = {'states': [{
            'name': "action",
            'tech': {
                'get': "/",
            }
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        with pytest.raises(CwsWriterError) as pytest_wrapped_e:
            sfn.generate()
        assert pytest_wrapped_e.type == CwsWriterError
        assert pytest_wrapped_e.value.args[0] == "The key service is missing for {'get': '/'}"

        data = {'states': [{
            'name': "action",
            'tech': {
                'service': "tech1",
                'get': "/",
                'post': "/",
            }
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        with pytest.raises(CwsWriterError) as pytest_wrapped_e:
            sfn.generate()
        assert pytest_wrapped_e.type == CwsWriterError
        assert pytest_wrapped_e.value.args[0].startswith("Too many methods defined for ")

    def test_catch_all(self):
        data = {'states': [{
            'name': "action",
            'tech': {
                'service': "tech1",
                'get': "/",
            }
        }], 'catch': [{'fail': None}]}
        sfn = TestStepFunction(yaml.dump(data))
        sfn.generate()
        assert len(sfn.all_states) == 4
        assert len(sfn.all_states[1].state['Catch']) == 1

        data = {'states': [{
            'name': "action",
            'tech': {
                'service': "tech1",
                'get': "/",
            },
            'catch': [{'fail': None}]
        }], 'catch': [{'fail': None}]}
        sfn = TestStepFunction(yaml.dump(data))
        sfn.generate()
        assert len(sfn.all_states) == 5
        assert len(sfn.all_states[1].state['Catch']) == 2

    def test_list(self):
        data = {'states': [{
            'name': "action 1",
            'tech': {
                'service': "tech",
                'get': "/",
            }
        }, {
            'name': "action 2",
            'tech': {
                'service': "tech",
                'get': "/",
            }
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        sfn.generate()
        assert 'Next' in sfn.all_states[1].state
        assert sfn.all_states[1].state['Next'] == sfn.all_states[2].name
        assert 'End' in sfn.all_states[2].state

    def test_choice(self):
        # missing key cases
        data = {'states': [{
            'name': "action",
            'choices': None
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        with pytest.raises(CwsWriterError):
            sfn.generate()

        data = {'states': [{
            'name': "action",
            'choices': [{
                'var': '', 'oper': '', 'value': '', 'goto': 'Step'
            }]
        }, {
            'name': "Step",
            'pass': None
        }]}
        sfn = TestStepFunction(yaml.dump(data))
        sfn.generate()
        assert len(sfn.all_states) == 3
        assert 'Default' in sfn.all_states[1].state
        assert 'End' in sfn.all_states[2].state



import pytest

from coworks.cws.client import client


class TestClass:

    def test_cmd_wront_project_dir(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'doesntexist', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    def test_cmd_no_service(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'quickstart2', 'cmd'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    def test_cmd_wrong_service(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'test', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    def test_cmd(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'tech_app', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == "test command with a=default/test command with b=value"

    def test_cmd_wrong_option(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'tech_app', 'test', '-t', 'wrong'],
                   obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    def test_cmd_right_option(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'tech_app', 'test', '-a', 'right'],
                   obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == "test command with a=right/test command with b=value"

    def test_cmd_wrong_b_option(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'tech_app', 'test', '-b', 'right'],
                   obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    def test_cmd_right_b_option(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'tech_app', 'test', '--b', 'right'],
                   obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == "test command with a=default/test command with b=right"

    def test_run(self, example_dir):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'info', 'run'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    def test_run_with_int_option(self, example_dir):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', example_dir, '-m', 'example', '-s', 'project1', 'run', '-p', '8000'],
                   obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0

    def test_export(self, example_dir):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws',
                   args=['-p', example_dir, '-m', 'example', '-s', 'tech_app', 'terraform', '--config', None], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0



import io
import threading
import time

import requests

from coworks.cws.runner import ThreadedLocalServer
from coworks.utils import import_attr


class TestClass:

    def test_run_quickstart2(self, example_dir):
        app = import_attr('quickstart2', 'app', cwd=example_dir)
        port = ThreadedLocalServer.unused_tcp_port()
        server = threading.Thread(target=run_server_quickstart, args=(app, port, example_dir), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{port}/', headers={'Authorization': "token"})
        assert response.text == "Simple microservice ready.\n"

    def test_export_quickstart2(self, example_dir):
        app = import_attr('quickstart2', 'app', cwd=example_dir)
        output = io.StringIO()
        app.execute('export', project_dir=example_dir, module='quickstart2', workspace='dev', output=output)
        output.seek(0)
        print(output.read())


def run_server_quickstart(app, port, example_dir):
    print(f"Server starting on port {port}")
    app.execute('run', host='localhost', port=port, project_dir=example_dir, module='quickstart2', workspace='dev')



import sys
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, asdict
from logging import getLogger, WARNING

import anyconfig
import click

from .error import CwsClientError
from ..config import DEFAULT_PROJECT_DIR, DEFAULT_WORKSPACE
from ..utils import import_attr, get_system_info
from ..version import __version__


@click.group()
@click.version_option(version=__version__, message=f'%(prog)s %(version)s, {get_system_info()}')
@click.option('-p', '--project-dir', default=DEFAULT_PROJECT_DIR,
              help='The project directory path (absolute or relative). Defaults to CWD')
@click.option('-m', '--module', help="Filename of your microservice python source file.")
@click.option('-s', '--service', help="Coworks application in the source file.")
@click.option('-w', '--workspace', default=DEFAULT_WORKSPACE, help="Application stage.")
@click.pass_context
def client(*args, **kwargs):
    ...


def invoke(ctx):
    """Invokes the command over the service or the declared services in project configuration file."""
    try:
        args = ctx.args
        protected_args = ctx.protected_args

        command_name = protected_args[0] if protected_args else None

        project_dir = ctx.params.get('project_dir')
        workspace = ctx.params.get('workspace')
        module = ctx.params.get('module')
        service = ctx.params.get('service')

        project_config = ProjectConfig(project_dir)
        if service:
            services = [(module, service)]
        else:
            services = project_config.all_services(module)

        if not services:
            sys.stderr.write(str("Nothing to execute as no service defined."))
            sys.exit(1)

        # Iterates over the declared services in project configuration file
        commands_to_be_executed = defaultdict(list)
        for module, service in services:
            ctx.args = list(args)
            ctx.protected_args = protected_args
            service_config = project_config.get_service_config(module, service, workspace)

            # Get command from the microservice
            handler = get_handler(project_dir, module, service)
            command = service_config.get_command(command_name, handler)
            if not command:
                raise CwsClientError(f"Undefined command {command_name}.\n")
            command_options = service_config.get_command_options(command_name)

            # Get user defined options and convert them in right types
            client_options, _, cmd_opts = command.make_parser(ctx).parse_args(ctx.args)
            for opt_key, opt_value in client_options.items():
                cmd_opt = next(x for x in cmd_opts if x.name == opt_key)
                client_options[opt_key] = cmd_opt.type(opt_value)

            execution_params = {**command_options, **client_options}
            command.make_context(command.name, execution_params)
            commands_to_be_executed[type(command)].append((command, execution_params))
        for command_class, execution_params in commands_to_be_executed.items():
            command_class.multi_execute(project_dir, workspace, execution_params)
    except CwsClientError as client_err:
        sys.stderr.write(client_err.msg)
        sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(1)


client.invoke = invoke


class ProjectConfig:
    """Class for the project configuration file."""

    def __init__(self, project_dir, file_name="cws.project", file_suffix=".yml"):
        from pathlib import Path
        self.project_dir = project_dir
        self.params = {}

        project_dir_path = Path(project_dir)
        self.project_file = project_dir_path / (file_name + file_suffix)
        project_secret_file = project_dir_path / (file_name + '.secret' + file_suffix)
        getLogger('anyconfig').setLevel(WARNING)
        self.params = anyconfig.multi_load([self.project_file, project_secret_file], ac_ignore_missing=True)

    def get_service_config(self, module, service, workspace):
        return ServiceConfig(self, module, service, workspace)

    def all_services(self, module=None):
        """ Returns the list of microservices on which the command will be executed."""
        services = self.params.get('services', {})

        res = []
        for s in services:
            if 'module' not in s:
                continue

            if module and s['module'] != module:
                continue

            if 'service' in s:
                res.append((s['module'], s['service']))
            elif 'services' in s:
                for ss in s['services']:
                    if 'service' in ss:
                        res.append((s['module'], ss['service']))
        return res

    @property
    def all_commands(self):
        """ Returns the list of microservices on which the command will be executed."""
        return self.params.get('commands', {})

    @staticmethod
    def _get_workspace_options(options, workspace):
        """Returns the option values defined for the specific workspace or globally."""
        workspaces = options.pop('workspaces', {})
        workspace_options = {k: v for x in workspaces if x.pop('workspace', None) == workspace
                             for k, v in x.items()}
        return {**options, **workspace_options}

    def _get_service_options(self, services, service, workspace):
        """Returns the option values defined for the specific service and workspace or globally."""

        service_options = {}
        for s in services:
            if s.pop('service', None) == service:
                s.pop('module', None)
                service_options.update(self._get_workspace_options(s, workspace))
        return {**service_options}

    def get_module_options(self, options_list, module, service, workspace):
        """Returns the option values defined for the specific module, service and workspace or globally."""

        if type(options_list) is not list:
            options_list = [options_list]

        service_options = {}
        module_options = {}
        for options in options_list:
            if 'module' not in options or options.pop('module') == module:
                services = options.pop('services', {})
                module_options.update(self._get_workspace_options(options, workspace))
                service_options.update(self._get_service_options(services, service, workspace))

        return {**module_options, **service_options}


@dataclass
class ServiceConfig:
    project_config: ProjectConfig
    module: str
    service: str
    workspace: str

    @property
    def client_params(self):
        res = asdict(self)
        del res['project_config']
        res['project_dir'] = self.project_config.project_dir
        return res

    def get_command(self, cmd_name, ms):
        """Get the command associated to this microservice."""

        # Get command already added in handler
        for name in ms.commands:
            if name == cmd_name:
                return ms.commands[name]

        # Creates it from project class parameter if not already defined
        cmd_class = self._command_class(cmd_name)
        if cmd_class:
            cmd = cmd_class(ms, name=cmd_name)

            # Installs needed commands
            for needed in cmd.needed_commands:
                self.get_command(needed, ms)

            click.help_option()(cmd)
            return cmd

    def _command_class(self, cmd_name):
        cmd_class_name = self.get_command_options(cmd_name).get('class')
        if cmd_class_name:
            splitted = cmd_class_name.split('.')
            return import_attr('.'.join(splitted[:-1]), splitted[-1], cwd=self.project_config.project_dir)

    def get_command_options(self, cmd_name):
        options = deepcopy(self.project_config.all_commands.get(cmd_name, {}))
        module_options = self.project_config.get_module_options(options, self.module, self.service, self.workspace)
        return {**self.client_params, **module_options}


def main():
    return client(obj={})


def get_handler(project_dir, module, service):
    # Load microservice handler
    try:
        return import_attr(module, service, cwd=project_dir)
    except AttributeError as e:
        raise CwsClientError(f"Module '{module}' has no microservice {service} : {str(e)}\n")
    except ModuleNotFoundError as e:
        raise CwsClientError(f"The module '{module}' is not defined in {project_dir} : {str(e)}\n")
    except Exception as e:
        raise CwsClientError(f"Error {e} when loading module '{module}'\n")


if __name__ == "__main__":
    main()



import contextlib
import logging
import os
import socket
import sys
import threading
import urllib.parse
from threading import Thread, Event

import chalice.local
import click
from aws_xray_sdk.core import xray_recorder
from chalice.config import Config
from chalice.local import ChaliceRequestHandler, LocalGateway, LocalDevServer

from .command import CwsCommand
from ..utils import threaded


class CwsRunner(CwsCommand):
    def __init__(self, app=None, name='run'):
        super().__init__(app, name=name)
        xray_recorder.configure(context_missing="LOG_ERROR")

    @property
    def options(self):
        return [
            *super().options,
            click.option('-h', '--host', default='127.0.0.1'),
            click.option('-p', '--port', default=8000, type=click.INT),
            click.option('--debug/--no-debug', default=False, help='Print debug logs to stderr.')
        ]

    @threaded
    def _execute(self, *, project_dir, workspace, host, port, debug, **options):
        """ Runs the microservice in a local Chalice emulator.

        :param host: the hostname to listen on.
        :param port: the port of the webserver.
        :param project_dir: to be able to import the microservice module.
        :param debug: if given, enable or disable debug mode.
        :param workspace: the workspace stagging run mode.
        :return: None
        """

        # chalice.cli package is not defined in deployment
        from .factory import CwsFactory

        ms = self.app
        os.environ['WORKSPACE'] = workspace
        ms.config.load_environment_variables(project_dir)
        if ms.entries is None:
            ms.deferred_init(workspace)

        if debug:
            logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(message)s')

        factory = CwsFactory(ms, project_dir, debug=debug)
        config = factory.mock_config_obj(ms)
        ms.local_server = LocalDevServer(ms, config, host, port, handler_cls=CwsRequestHandler)
        ms.__class__ = type('LocalMicroService', (ms.__class__, ThreadedMixin), {})
        ms.local_server.serve_forever()


class CwsRequestHandler(ChaliceRequestHandler):
    """Request handler redefined to quote query parameters."""

    def __init__(self, request, client_address, server, app_object, config):
        chalice.local.LocalGateway = CwsLocalGateway
        super().__init__(request, client_address, server, app_object, config)
        chalice.local.LocalGateway = LocalGateway

    def parse_request(self):
        request = super().parse_request()
        self.path = urllib.parse.quote(self.path)
        return request


class CwsLocalGateway(LocalGateway):
    """Local gateway redefined to unquote query parameters."""

    def _generate_lambda_event(self, method, path, headers, body):
        path = urllib.parse.unquote(path)
        return super()._generate_lambda_event(method, path, headers, body)


class ThreadedMixin:
    _THREAD_LOCAL = threading.local()

    @property
    def current_request(self):
        return self._THREAD_LOCAL.current_request

    @current_request.setter
    def current_request(self, value):
        self._THREAD_LOCAL.current_request = value


class ThreadedLocalServer(Thread):
    threaded_servers = {}

    def __init__(self, *, port=None, host='localhost'):
        super().__init__()
        self._app_object = None
        self._config = None
        self._host = host
        self._port = port or self.unused_tcp_port()
        self._server = None
        self._server_ready = Event()

    def wait_for_server_ready(self):
        self._server_ready.wait()

    def configure(self, app_object, config=None, **kwargs):
        self._app_object = app_object
        self._config = config if config else Config()

    def run(self):
        self._server = LocalDevServer(self._app_object, self._config, self._host, self._port)
        self._server_ready.set()
        self._server.serve_forever()

    def make_call(self, method, path, timeout=0.5, **kwarg):
        self._server_ready.wait()
        return method('http://{host}:{port}{path}'.format(
            path=path, host=self._host, port=self._port), timeout=timeout, **kwarg)

    def shutdown(self):
        if self._server is not None:
            self._server.server.shutdown()

    @classmethod
    def unused_tcp_port(cls):
        with contextlib.closing(socket.socket()) as sock:
            sock.bind(('localhost', 0))
            return sock.getsockname()[1]



import json
import os
import urllib.parse
import tempfile

import requests

from coworks.tech.jinja import JinjaRenderMicroService


class TestClass:

    def test_render_template_in_url(self, local_server_factory):
        local_server = local_server_factory(JinjaRenderMicroService())
        template = urllib.parse.quote_plus("hello {{ world_name }}")
        response = local_server.make_call(requests.get, f"/render/{template}", params={'world_name': 'world'})
        assert response.status_code == 200
        assert response.json() == {"render": "hello [\'world\']"}

    def test_render_template_multipart_form(self, local_server_factory):
        local_server = local_server_factory(JinjaRenderMicroService())

        template = tempfile.TemporaryFile()
        template.write("hello {{ world_name }}".encode())
        template.seek(0)

        context = tempfile.TemporaryFile()
        context.write(json.dumps({"world_name": "the world"}).encode())
        context.seek(0)

        response = local_server.make_call(requests.post, f"/render/template.jinja",
                                          files={'templates': ('template.jinja', template, 'text/plain'),
                                                 'context': (None, context, 'application/json')})
        assert response.status_code == 200
        assert response.json() == {"render": "hello the world"}



import os

from setuptools import setup, find_packages

_here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

version = {}
with open(os.path.join(_here, 'coworks', 'version.py')) as f:
    exec(f.read(), version)

setup(
    name='coworks',
    version=version['__version__'],
    description='Coworks is a unified compositional microservices framework over AWS serverless technologies.',
    long_description=long_description,
    author='Guillaume Doumenc',
    author_email='gdoumenc@fpr-coworks.com',
    url='https://github.com/gdoumenc/coworks',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'anyconfig>=0.9.11',
        'aws_xray_sdk>=2.5.0',
        'boto3>=1.13.15',
        'chalice>=1.14.1',
        'jinja2>=2.11',
        'pyyaml>=5.3.1',
        'requests_toolbelt>=0.9.1',
        'python_terraform>=0.10.1'
    ],
    keywords='python3 serverless microservice aws-lambda aws step-functions',
    entry_points={
        'console_scripts': ['cws=coworks.cws.client:main'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
        'Topic :: System :: Distributed Computing'
    ],
)



import pytest

from coworks.cws.client import client, ProjectConfig


class TestClass:
    def test_no_project_file_no_module(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1

    def test_project_file_no_param(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests/example', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == \
               "test command with a=default/test command with b=valuetest command with a=project2/test command with b=value"

    def test_project_file_no_param_workspace_dev(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests/example', '-w', 'dev', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == \
               "test command with a=default/test command with b=valuetest command with a=project2/test command with b=value"

    def test_project_file_no_param_workspace_prod(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests/example', '-w', 'prod', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == \
               "test command with a=default/test command with b=valuetest command with a=prod2/test command with b=value"

    def test_project_file_param(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests/example', 'test', '-a', 'param'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == \
               "test command with a=param/test command with b=valuetest command with a=param/test command with b=value"

    def test_project_file_project1(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests/example', '-m', 'example', '-s', 'project1', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == "test command with a=default/test command with b=value"

    def test_project_file_project2(self, example_dir, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'tests/example', '-m', 'example', '-s', 'project2', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
        captured = capsys.readouterr()
        assert captured.out == "test command with a=project2/test command with b=value"

    def test_all_services(self):
        conf = ProjectConfig('tests/src/cws')
        services = conf.all_services()
        assert len(services) == 3
        assert ('m1', 's1') in services
        assert ('m1', 's2') in services
        assert ('m1', 's3') not in services
        assert ('m2', 's3') in services

    def test_command0(self):
        conf = ProjectConfig('tests/src/cws')
        service_conf = conf.get_service_config('module', 'service', 'dev')
        commands = conf.all_commands
        assert len(commands) == 4
        options = service_conf.get_command_options('test_command0')
        assert len(options) == 7
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value2"
        assert options['option_key3'] == "option_value3"

    def test_command1(self):
        conf = ProjectConfig('tests/src/cws')
        service_conf = conf.get_service_config('module', 'service', 'dev')
        options = service_conf.get_command_options('test_command1')
        assert len(options) == 7
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value2"
        assert options['option_key3'] == "option_value5"
        service_conf = conf.get_service_config('module', 'service', 'prod')
        options = service_conf.get_command_options('test_command1')
        assert len(options) == 7
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value4"
        assert options['option_key3'] == "option_value5"

    def test_command2_no_service(self):
        conf = ProjectConfig('tests/src/cws')
        service_conf = conf.get_service_config('module', 'service', 'dev')
        options = service_conf.get_command_options('test_command2')
        assert len(options) == 6
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value2"

    def test_command2_wrong_service(self):
        conf = ProjectConfig('tests/src/cws')
        service_conf = conf.get_service_config('m1', 'service', 'dev')
        options = service_conf.get_command_options('test_command2')
        assert len(options) == 7
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value4"
        assert options['option_key4'] == "option_value8"
        service_conf = conf.get_service_config('m1', 'service', 'prod')
        options = service_conf.get_command_options('test_command2')
        assert len(options) == 6
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value4"

    def test_command2_service_s1(self):
        conf = ProjectConfig('tests/src/cws')
        service_conf = conf.get_service_config('m1', 's1', 'dev')
        options = service_conf.get_command_options('test_command2')
        assert len(options) == 8
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value4"
        assert options['option_key3'] == "option_value5"
        assert options['option_key4'] == "option_value8"
        service_conf = conf.get_service_config('m1', 's1', 'prod')
        options = service_conf.get_command_options('test_command2')
        assert len(options) == 7
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value4"
        assert options['option_key3'] == "option_value5"
        assert 'option_key4' not in options

    def test_command2_service_s2(self):
        conf = ProjectConfig('tests/src/cws')
        service_conf = conf.get_service_config('m1', 's2', 'dev')
        options = service_conf.get_command_options('test_command2')
        assert len(options) == 7
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value4"
        assert options['option_key4'] == "option_value7"
        service_conf = conf.get_service_config('m1', 's2', 'prod')
        options = service_conf.get_command_options('test_command2')
        assert len(options) == 7
        assert options['option_key1'] == "option_value1"
        assert options['option_key2'] == "option_value4"
        assert options['option_key4'] == "option_value6"



import requests
from unittest.mock import Mock

from tests.src.coworks.blueprint.blueprint import BP
from tests.src.coworks.tech_ms import SimpleMS


class TestClass:
    def test_request(self, local_server_factory):
        ms = SimpleMS()
        ms.register_blueprint(BP(import_name="blueprint"))
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == 'get'
        response = local_server.make_call(requests.get, '/test/3')
        assert response.status_code == 200
        assert response.text == 'blueprint test 3'
        response = local_server.make_call(requests.get, '/extended/test/3')
        assert response.status_code == 200
        assert response.text == 'blueprint extended test 3'

    def test_prefix(self, local_server_factory):
        ms = SimpleMS()
        ms.register_blueprint(BP(), url_prefix="/prefix")
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.get, '/prefix/test/3')
        assert response.status_code == 200
        assert response.text == 'blueprint test 3'
        response = local_server.make_call(requests.get, '/prefix/extended/test/3')
        assert response.status_code == 200
        assert response.text == 'blueprint extended test 3'

    def test_before_activation(self, local_server_factory):
        ms = SimpleMS()
        init_bp = InitBP()
        ms.register_blueprint(init_bp, url_prefix="/prefix")
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.get, '/prefix/test/3')
        assert response.status_code == 200
        assert response.text == 'blueprint test 3'
        init_bp.do_before_first_activation.assert_called_once()
        init_bp.do_before_activation.assert_called_once()
        init_bp.do_after_activation.assert_called_once()


class InitBP(BP):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.do_before_first_activation = Mock()
        self.do_before_activation = Mock()
        self.do_after_activation = Mock()

        @self.before_first_activation
        def before_first_activation(event, context):
            self.do_before_first_activation(event, context)

        @self.before_activation
        def before_activation(event, context):
            self.do_before_activation(event, context)

        @self.after_activation
        def after_activation(response):
            self.do_after_activation(response)
            return response



import requests

from tests.src.coworks.tech_ms import *


class TestClass:

    def test_request_arg(self, local_server_factory):
        local_server = local_server_factory(SimpleMS())
        response = local_server.make_call(requests.get, '/')
        assert response.status_code == 200
        assert response.text == 'get'
        response = local_server.make_call(requests.post, '/')
        assert response.status_code == 405
        response = local_server.make_call(requests.get, '/get1')
        assert response.status_code == 403
        response = local_server.make_call(requests.get, '/content')
        assert response.status_code == 200
        assert response.text == "get content"
        response = local_server.make_call(requests.get, '/content/3')
        assert response.status_code == 200
        assert response.text == "get content with 3"
        response = local_server.make_call(requests.get, '/content/3/other')
        assert response.status_code == 200
        assert response.text == "get content with 3 and other"
        response = local_server.make_call(requests.post, '/content', json={"other": 'other'})
        assert response.status_code == 200
        assert response.text == "post content without value but other"
        response = local_server.make_call(requests.post, '/content/3', json={"other": 'other'})
        assert response.status_code == 200
        assert response.text == "post content with 3 and other"
        response = local_server.make_call(requests.post, '/content/3', json='other')
        assert response.status_code == 200
        assert response.text == "post content with 3 and other"
        response = local_server.make_call(requests.post, '/content/3', json={"other": 'other', "value": 5})
        assert response.status_code == 400

    def test_request_kwargs(self, local_server_factory):
        local_server = local_server_factory(SimpleMS())
        response = local_server.make_call(requests.get, '/kwparam1', params={"value": 5})
        assert response.status_code == 200
        assert response.text == "get **param with only 5"
        response = local_server.make_call(requests.get, '/kwparam1', params={"other": 'other', "value": 5})
        assert response.status_code == 400
        response = local_server.make_call(requests.get, '/kwparam1', json={"value": 5})
        assert response.status_code == 200
        assert response.text == "get **param with only 5"
        response = local_server.make_call(requests.get, '/kwparam1', json={"other": 'other', "value": 5})
        assert response.status_code == 400
        response = local_server.make_call(requests.get, '/kwparam2', params={"other": 'other', "value": 5})
        assert response.status_code == 200
        assert response.text == "get **param with 5 and ['other']"
        response = local_server.make_call(requests.get, '/kwparam2', json={"other": 'other', "value": 5})
        assert response.status_code == 200
        assert response.text == "get **param with 5 and ['other']"

        response = local_server.make_call(requests.get, '/extended/content')
        assert response.status_code == 200
        assert response.text == "hello world"

    def test_parameterized(self, local_server_factory):
        local_server = local_server_factory(ParamMS())
        response = local_server.make_call(requests.get, '/123')
        assert response.status_code == 200
        assert response.text == '123'
        response = local_server.make_call(requests.get, '/concat/123/456')
        assert response.status_code == 200
        assert response.text == '123456'
        response = local_server.make_call(requests.get, '/value')
        assert response.status_code == 200
        assert response.text == '123'
        response = local_server.make_call(requests.put, '/value', json={'value': "456"})
        assert response.status_code == 200
        assert response.text == '456'
        response = local_server.make_call(requests.get, '/value')
        assert response.status_code == 200
        assert response.text == '456'
        response = local_server.make_call(requests.get, '/param/test1')
        assert response.status_code == 200
        assert response.text == 'test1default1default2'
        response = local_server.make_call(requests.get, '/param/test1', params={"param1": "param1"})
        assert response.status_code == 200
        assert response.text == 'test1param1default2'
        response = local_server.make_call(requests.get, '/param/test1', params={"param2": "param2"})
        assert response.status_code == 200
        assert response.text == 'test1default1param2'
        response = local_server.make_call(requests.get, '/param/test1', params={"param1": "param1", "param2": "param2"})
        assert response.status_code == 200
        assert response.text == 'test1param1param2'
        response = local_server.make_call(requests.get, '/param/test1',
                                          params=[('param1', 'param1'), ('param1', 'param2')])
        assert response.status_code == 200
        assert response.text == "test1['param1', 'param2']default2"

    def test_slug_parameterized(self, local_server_factory):
        local_server = local_server_factory(ParamMS())
        response = local_server.make_call(requests.get, '/123')
        assert response.status_code == 200
        assert response.text == '123'
        response = local_server.make_call(requests.get, '/concat/123/456')
        assert response.status_code == 200
        assert response.text == '123456'

    def test_tuple_returned(self, local_server_factory):
        local_server = local_server_factory(TupleReturnedMS())
        headers = {'Content-type': 'text/plain', 'Accept': 'text/plain'}
        response = local_server.make_call(requests.get, '/', headers=headers)
        assert response.status_code == 200
        assert response.text == 'ok'
        assert response.headers['content-type'] == 'application/json'
        response = local_server.make_call(requests.get, '/json')
        assert response.status_code == 200
        assert response.json()['value'] == 'ok'
        assert response.headers['content-type'] == 'application/json'
        response = local_server.make_call(requests.get, '/resp/ok')
        assert response.status_code == 200
        assert response.text == 'ok'
        assert response.headers['content-type'] == 'application/json'
        response = local_server.make_call(requests.get, '/tuple/test')
        assert response.status_code == 200
        assert response.headers['content-type'] == 'application/json'
        assert response.headers['x-test'] == 'true'
        assert response.text == 'test'




import sys
from abc import ABC, abstractmethod

import click

from coworks import TechMicroService
from coworks.cws.error import CwsCommandError


class CwsCommand(click.Command, ABC):

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_params):
        for command, options in execution_params:
            command.execute(**options)

    def __init__(self, app: TechMicroService = None, *, name):
        super().__init__(name, callback=self._execute)

        # Trace interfaces.
        self.output = sys.stdout
        self.error = sys.stderr

        # A list of functions that will be called before or after the command executioon.
        self.before_funcs = []
        self.after_funcs = []

        if app is not None:
            self.app = app
            self.init_app(app)

        for opt in self.options:
            opt(self)

    def init_app(self, app):
        app.commands[self.name] = self
        for cmd in self.needed_commands:
            if cmd not in app.commands:
                raise CwsCommandError(f"Undefined command {cmd} needed.")

            for opt in self.app.commands[cmd].options:
                opt(self)

    @property
    def needed_commands(self):
        return []

    @property
    def options(self):
        return []

    def execute(self, *, project_dir, module, service, workspace, output=None, error=None, **options):
        """ Called when the command is called.
        :param output: output stream.
        :param error: error stream.
        :param options: command options.
        :return: None
        """
        self.app.deferred_init(workspace)

        if output is not None:
            self.output = open(output, 'w+') if type(output) is str else output
        if error is not None:
            self.error = open(error, 'w+') if type(error) is str else error

        try:
            for func in self.before_funcs:
                func(options)

            ctx = self.make_context(self.name, options)
            ctx_options = {**options, 'output': output, 'error': error}
            ctx.params.update(project_dir=project_dir, module=module, service=service, workspace=workspace,
                              **ctx_options)
            self.invoke(ctx)

            for func in self.after_funcs:
                func(options)
        except click.exceptions.Exit as e:
            if e.exit_code:
                raise CwsCommandError("Command exits with error")
            return
        except CwsCommandError:
            raise
        except Exception as e:
            raise CwsCommandError(str(e))

    def before_execute(self, f):
        """Registers a function to be run before the command execution.
        :param f: function called before the command execution
        :return: None

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.before_funcs.append(f)
        return f

    def after_execute(self, f):
        """Registers a function to be run after the command execution.
        :param f: function called after the command execution
        :return: None

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.after_funcs.append(f)
        return f

    def parse_args(self, ctx, args):
        for param in self.get_params(ctx):
            if param.name not in args:
                if param.required:
                    raise CwsCommandError(f"missing parameter: {param.name}")
                args[param.name] = param.get_default(ctx)
        ctx.args = args
        return args

    @abstractmethod
    def _execute(self, **options):
        """ Main command function.
        :param options: Command options.
        :return: None.

        Abstract method which must be redefined in any subclass. The content should be written in self.output.
        """



import json

import requests

from coworks import Blueprint
from coworks.blueprint import Admin
from tests.src.coworks.tech_ms import *


class DocumentedMS(TechMS):

    def get(self):
        """Root access."""
        return "get"

    def post_content(self, value, other="none"):
        """Add content."""
        return f"post_content {value}{other}"

    def post_contentannotated(self, value: int, other: str = "none"):
        """Add content."""
        return f"post_content {value}{other}"

    def get_list(self, values: [int]):
        """Tests list param."""
        return "ok"


class HiddenBlueprint(Blueprint):

    def get(self):
        """Test not in routes."""
        return "ok"


class TestClass:
    def test_documentation(self, local_server_factory):
        ms = DocumentedMS()
        ms.register_blueprint(Admin(), url_prefix="/admin")
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.get, '/admin/routes', timeout=500)
        assert response.status_code == 200
        routes = json.loads(response.text)
        assert routes["/"] == {
            "GET": {
                "doc": "Root access.",
                "signature": "()"
            }
        }
        assert routes["/content/{_0}"] == {
            "POST": {
                "doc": "Add content.",
                "signature": "(value, other=none)"
            }
        }
        assert routes["/contentannotated/{_0}"] == {
            "POST": {
                "doc": "Add content.",
                "signature": "(value:<class 'int'>, other:<class 'str'>=none)"
            }
        }
        assert routes["/admin/routes"] == {
            "GET": {
                "doc": 'Returns the list of entrypoints with signature.',
                "signature": "()"
            }
        }
        assert routes["/list/{_0}"] == {
            "GET": {
                "doc": 'Tests list param.',
                "signature": "(values:[<class 'int'>])"
            }
        }

    def test_documentation(self, local_server_factory):
        ms = DocumentedMS()
        ms.register_blueprint(Admin(), url_prefix="/admin")
        ms.register_blueprint(HiddenBlueprint(), url_prefix="/hidden", hide_routes=True)
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.get, '/admin/routes', timeout=500)
        assert response.status_code == 200
        routes = json.loads(response.text)
        assert '/hidden' not in routes



import requests
from chalice import AuthResponse

from coworks.config import Config
from .blueprint.blueprint import BP
from .tech_ms import SimpleMS


class AuthorizeAll(SimpleMS):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.register_blueprint(BP(), url_prefix="/blueprint")

    def auth(self, auth_request):
        return True

    def get_product(self, ref):
        return ref

class AuthorizeNothingExceptBP(SimpleMS):

    def __init__(self):
        super().__init__()
        self.register_blueprint(BP(), url_prefix="/blueprint", authorizer=auth_external)

    def auth(self, auth_request):
        return False


class AuthorizeNothing(SimpleMS):

    def __init__(self):
        super().__init__()
        self.register_blueprint(BP(), url_prefix="/blueprint")

    def auth(self, auth_request):
        return False


class AuthorizedMS(SimpleMS):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.register_blueprint(BP(), url_prefix="/blueprint")

    def auth(self, auth_request):
        if auth_request.token == 'allow':
            return AuthResponse(routes=['*'], principal_id='user')
        return False


class AuthorizedMS2(SimpleMS):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.register_blueprint(BP(), url_prefix="/blueprint", authorizer=AuthorizedMS2.auth)

    def auth(self, auth_request):
        if auth_request.token == 'allow':
            return AuthResponse(routes=['/'], principal_id='user')
        return False


class TestClass:
    def test_authorize_all(self, local_server_factory):
        local_server = local_server_factory(AuthorizeAll())
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'token'})
        assert response.status_code == 200
        assert response.text == 'get'
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'token'})
        assert response.status_code == 200
        assert response.text == 'blueprint test 3'

    def test_authorize_nothing_except_bp(self, local_server_factory):
        local_server = local_server_factory(AuthorizeNothingExceptBP())
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
        assert response.status_code == 403
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'allow'})
        assert response.status_code == 200

    def test_authorize_nothing(self, local_server_factory):
        local_server = local_server_factory(AuthorizeNothing())
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
        assert response.status_code == 403
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'allow'})
        assert response.status_code == 403

    def test_authorized(self, local_server_factory):
        local_server = local_server_factory(AuthorizedMS())
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
        assert response.status_code == 200
        assert response.text == 'get'
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'refuse'})
        assert response.status_code == 403
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'allow'},
                                          timeout=500)
        assert response.status_code == 200
        assert response.text == 'blueprint test 3'
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'refuse'})
        assert response.status_code == 403

    def test_authorized2(self, local_server_factory):
        local_server = local_server_factory(AuthorizedMS2())
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
        assert response.status_code == 200
        assert response.text == 'get'
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'refuse'})
        assert response.status_code == 403
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'allow'})
        assert response.status_code == 403
        assert response.text == '{"Message": "User is not authorized to access this resource"}'
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'refuse'})
        assert response.status_code == 403
        assert response.text == '{"Message": "User is not authorized to access this resource"}'

    def test_authorized_external(self, local_server_factory):
        config = Config(auth=auth_external)
        local_server = local_server_factory(AuthorizedMS(configs=[config]))
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
        assert response.status_code == 200
        assert response.text == 'get'
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'refuse'})
        assert response.status_code == 403
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'allow'})
        assert response.status_code == 200
        assert response.text == 'blueprint test 3'
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'refuse'})
        assert response.status_code == 403

    def test_auth_entry(self, local_server_factory):
        ms = AuthorizedMS()
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
        assert response.status_code == 200
        assert ms.entry('/').authorizer
        assert response.status_code == 200
        response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'allow'})
        assert response.status_code == 200
        assert ms.entry('/blueprint/test/3').authorizer

    def test_entries(self, local_server_factory):
        ms = AuthorizeAll()
        local_server = local_server_factory(ms)
        response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
        assert response.status_code == 200
        assert ms.entry('wrong') is None
        assert ms.entry('content') is not None
        assert ms.entry('content/value') is not None
        assert ms.entry('content/value/other') is not None
        assert ms.entry('content/value/other/wrong') is None
        assert ms.entry('/product/TRANSPORT') is not None


def auth_external(self, auth_request):
    return True if auth_request.token == 'allow' else False



from pathlib import Path

import click
from python_terraform import Terraform

from coworks.cws.command import CwsCommand, CwsCommandError


class CwsInformant(CwsCommand):
    """Command to get information on the project's microservices and teir deployment."""

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_params):
        terraform, tf_dir = False, ''
        for command, options in execution_params:
            command.print_module_info(**options)

            terraform, tf_dir = options['terraform'], options['tf_dir']

            if options['env']:
                command.print_env_vars(**options)

        if terraform:
            cls.print_terraform_output(tf_dir)

    def __init__(self, app=None, name='info'):
        super().__init__(app, name=name)

    @property
    def options(self):
        return [
            *super().options,
            click.option('--env', '-e', is_flag=True, help="Show environment variables."),
            click.option('--terraform', '-t', is_flag=True, help="Show deployed terraform output."),
            click.option('--tf_dir', default='terraform', help="Terraform directory."),
        ]

    def _execute(self, *, workspace, **options):
        raise CwsCommandError("Not implemented")

    def print_module_info(self, *, module, service, **options):
        print(f"Microservice {service} defined in module {module}")

    def print_env_vars(self, *, project_dir, workspace, service, **options):
        for config in self.app.configs:
            if config.workspace == workspace:
                print(f"Environment vars for {service} in workspace {config.workspace}:")
                files = config.existing_environment_variables_files(project_dir)
                for file in files:
                    with file.open() as f:
                        print(f.read())

    @classmethod
    def print_terraform_output(cls, tf_dir):
        if Path(tf_dir).exists():
            terraform = Terraform(tf_dir)
            print(terraform.output())
